# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests for composer.Entity."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import itertools
# Internal dependencies.

from absl.testing import absltest
from absl.testing import parameterized
from dm_control import mjcf
from dm_control.composer import arena
from dm_control.composer import define
from dm_control.composer import entity
from dm_control.composer.observation.observable import base as observable
import numpy as np
import six
from six.moves import range

_NO_ROTATION = (1, 0, 0, 0)  # Tests support for non-arrays and non-floats.
_NINETY_DEGREES_ABOUT_X = np.array(
    [np.cos(np.pi / 4), np.sin(np.pi / 4), 0., 0.])
_NINETY_DEGREES_ABOUT_Y = np.array(
    [np.cos(np.pi / 4), 0., np.sin(np.pi / 4), 0.])
_NINETY_DEGREES_ABOUT_Z = np.array(
    [np.cos(np.pi / 4), 0., 0., np.sin(np.pi / 4)])
_FORTYFIVE_DEGREES_ABOUT_X = np.array(
    [np.cos(np.pi / 8), np.sin(np.pi / 8), 0., 0.])

_TEST_ROTATIONS = [
    # Triplets of original rotation, new rotation and final rotation.
    (None, _NO_ROTATION, _NO_ROTATION),
    (_NO_ROTATION, _NINETY_DEGREES_ABOUT_Z, _NINETY_DEGREES_ABOUT_Z),
    (_FORTYFIVE_DEGREES_ABOUT_X, _NINETY_DEGREES_ABOUT_Y,
     np.array([0.65328, 0.2706, 0.65328, -0.2706])),
]


def _param_product(**param_lists):
  keys, values = zip(*param_lists.items())
  for combination in itertools.product(*values):
    yield dict(zip(keys, combination))


class TestEntity(entity.Entity):
  """Simple test entity that does nothing but declare some observables."""

  def _build(self, name='test_entity'):
    self._mjcf_root = mjcf.element.RootElement(model=name)
    self._mjcf_root.worldbody.add('geom', type='sphere', size=(0.1,))

  def _build_observables(self):
    return TestEntityObservables(self)

  @property
  def mjcf_model(self):
    return self._mjcf_root


class TestEntityObservables(entity.Observables):
  """Trivial observables for the test entity."""

  @define.observable
  def observable0(self):
    return observable.Generic(lambda phys: 0.0)

  @define.observable
  def observable1(self):
    return observable.Generic(lambda phys: 1.0)


class EntityTest(parameterized.TestCase):

  def setUp(self):
    super(EntityTest, self).setUp()
    self.entity = TestEntity()

  def testNumObservables(self):
    """Tests that the observables dict has the right number of entries."""
    self.assertLen(self.entity.observables.as_dict(), 2)

  def testObservableNames(self):
    """Tests that the observables dict keys correspond to the observable names.
    """
    obs = self.entity.observables.as_dict()
    self.assertIn('observable0', obs)
    self.assertIn('observable1', obs)

    subentity = TestEntity(name='subentity')
    self.entity.attach(subentity)
    self.assertIn('subentity/observable0', subentity.observables.as_dict())
    self.assertEqual(subentity.observables.dict_keys.observable0,
                     'subentity/observable0')
    self.assertIn('observable0', dir(subentity.observables.dict_keys))
    self.assertIn('subentity/observable1', subentity.observables.as_dict())
    self.assertEqual(subentity.observables.dict_keys.observable1,
                     'subentity/observable1')
    self.assertIn('observable1', dir(subentity.observables.dict_keys))

  def testEnableDisableObservables(self):
    """Test the enabling and disable functionality for observables."""
    all_obs = self.entity.observables.as_dict()

    self.entity.observables.enable_all()
    for obs in all_obs.values():
      self.assertTrue(obs.enabled)

    self.entity.observables.disable_all()
    for obs in all_obs.values():
      self.assertFalse(obs.enabled)

    self.entity.observables.observable0.enabled = True
    self.assertTrue(all_obs['observable0'].enabled)

  def testObservableDefaultOptions(self):
    corruptor = lambda x: x
    options = {
        'update_interval': 2,
        'buffer_size': 10,
        'delay': 1,
        'aggregator': 'max',
        'corruptor': corruptor,
        'enabled': True
    }
    self.entity.observables.set_options(options)

    for obs in self.entity.observables.as_dict().values():
      self.assertEqual(obs.update_interval, 2)
      self.assertEqual(obs.delay, 1)
      self.assertEqual(obs.buffer_size, 10)
      self.assertEqual(obs.aggregator, observable.AGGREGATORS['max'])
      self.assertEqual(obs.corruptor, corruptor)
      self.assertTrue(obs.enabled)

  def testObservablePartialDefaultOptions(self):
    options = {'update_interval': 2, 'delay': 1}
    self.entity.observables.set_options(options)

    for obs in self.entity.observables.as_dict().values():
      self.assertEqual(obs.update_interval, 2)
      self.assertEqual(obs.delay, 1)
      self.assertEqual(obs.buffer_size, None)
      self.assertEqual(obs.aggregator, None)
      self.assertEqual(obs.corruptor, None)

  def testObservableOptionsInvalidName(self):
    options = {'asdf': None}
    with six.assertRaisesRegex(
        self, KeyError, 'No observable with name \'asdf\''):
      self.entity.observables.set_options(options)

  def testObservableInvalidOptions(self):
    options = {'observable0': {'asdf': 2}}
    with six.assertRaisesRegex(self, AttributeError,
                               'Cannot add attribute asdf in configure.'):
      self.entity.observables.set_options(options)

  def testObservableOptions(self):
    options = {
        'observable0': {
            'update_interval': 2,
            'delay': 3
        },
        'observable1': {
            'update_interval': 4,
            'delay': 5
        }
    }
    self.entity.observables.set_options(options)
    observables = self.entity.observables.as_dict()
    self.assertEqual(observables['observable0'].update_interval, 2)
    self.assertEqual(observables['observable0'].delay, 3)
    self.assertEqual(observables['observable0'].buffer_size, None)
    self.assertEqual(observables['observable0'].aggregator, None)
    self.assertEqual(observables['observable0'].corruptor, None)
    self.assertFalse(observables['observable0'].enabled)

    self.assertEqual(observables['observable1'].update_interval, 4)
    self.assertEqual(observables['observable1'].delay, 5)
    self.assertEqual(observables['observable1'].buffer_size, None)
    self.assertEqual(observables['observable1'].aggregator, None)
    self.assertEqual(observables['observable1'].corruptor, None)
    self.assertFalse(observables['observable1'].enabled)

  def testObservableOptionsEntityConstructor(self):
    options = {
        'observable0': {
            'update_interval': 2,
            'delay': 3
        },
        'observable1': {
            'update_interval': 4,
            'delay': 5
        }
    }
    ent = TestEntity(observable_options=options)
    observables = ent.observables.as_dict()
    self.assertEqual(observables['observable0'].update_interval, 2)
    self.assertEqual(observables['observable0'].delay, 3)
    self.assertEqual(observables['observable0'].buffer_size, None)
    self.assertEqual(observables['observable0'].aggregator, None)
    self.assertEqual(observables['observable0'].corruptor, None)
    self.assertFalse(observables['observable0'].enabled)

    self.assertEqual(observables['observable1'].update_interval, 4)
    self.assertEqual(observables['observable1'].delay, 5)
    self.assertEqual(observables['observable1'].buffer_size, None)
    self.assertEqual(observables['observable1'].aggregator, None)
    self.assertEqual(observables['observable1'].corruptor, None)
    self.assertFalse(observables['observable1'].enabled)

  def testObservablePartialOptions(self):
    options = {'observable0': {'update_interval': 2, 'delay': 3}}
    self.entity.observables.set_options(options)
    observables = self.entity.observables.as_dict()
    self.assertEqual(observables['observable0'].update_interval, 2)
    self.assertEqual(observables['observable0'].delay, 3)
    self.assertEqual(observables['observable0'].buffer_size, None)
    self.assertEqual(observables['observable0'].aggregator, None)
    self.assertEqual(observables['observable0'].corruptor, None)
    self.assertFalse(observables['observable0'].enabled)

    self.assertEqual(observables['observable1'].update_interval, 1)
    self.assertEqual(observables['observable1'].delay, None)
    self.assertEqual(observables['observable1'].buffer_size, None)
    self.assertEqual(observables['observable1'].aggregator, None)
    self.assertEqual(observables['observable1'].corruptor, None)
    self.assertFalse(observables['observable1'].enabled)

  def testAttach(self):
    entities = [TestEntity() for _ in range(4)]
    entities[0].attach(entities[1])
    entities[1].attach(entities[2])
    entities[0].attach(entities[3])

    self.assertIsNone(entities[0].parent)
    self.assertIs(entities[1].parent, entities[0])
    self.assertIs(entities[2].parent, entities[1])
    self.assertIs(entities[3].parent, entities[0])

    self.assertIsNone(entities[0].mjcf_model.parent_model)
    self.assertIs(entities[1].mjcf_model.parent_model, entities[0].mjcf_model)
    self.assertIs(entities[2].mjcf_model.parent_model, entities[1].mjcf_model)
    self.assertIs(entities[3].mjcf_model.parent_model, entities[0].mjcf_model)

    self.assertEqual(list(entities[0].iter_entities()), entities)

  def testDetach(self):
    entities = [TestEntity() for _ in range(4)]
    entities[0].attach(entities[1])
    entities[1].attach(entities[2])
    entities[0].attach(entities[3])

    entities[1].detach()
    with six.assertRaisesRegex(self, RuntimeError, 'not attached'):
      entities[1].detach()

    self.assertIsNone(entities[0].parent)
    self.assertIsNone(entities[1].parent)
    self.assertIs(entities[2].parent, entities[1])
    self.assertIs(entities[3].parent, entities[0])

    self.assertIsNone(entities[0].mjcf_model.parent_model)
    self.assertIsNone(entities[1].mjcf_model.parent_model)
    self.assertIs(entities[2].mjcf_model.parent_model, entities[1].mjcf_model)
    self.assertIs(entities[3].mjcf_model.parent_model, entities[0].mjcf_model)

    self.assertEqual(list(entities[0].iter_entities()),
                     [entities[0], entities[3]])

  def testIterEntitiesExcludeSelf(self):
    entities = [TestEntity() for _ in range(4)]
    entities[0].attach(entities[1])
    entities[1].attach(entities[2])
    entities[0].attach(entities[3])
    self.assertEqual(
        list(entities[0].iter_entities(exclude_self=True)), entities[1:])

  def testGlobalVectorToLocalFrame(self):
    parent = TestEntity()
    parent.mjcf_model.worldbody.add(
        'site', xyaxes=[0, 1, 0, -1, 0, 0]).attach(self.entity.mjcf_model)
    physics = mjcf.Physics.from_mjcf_model(parent.mjcf_model)

    # 3D vectors
    np.testing.assert_allclose(
        self.entity.global_vector_to_local_frame(physics, [0, 1, 0]),
        [1, 0, 0], atol=1e-10)
    np.testing.assert_allclose(
        self.entity.global_vector_to_local_frame(physics, [-1, 0, 0]),
        [0, 1, 0], atol=1e-10)
    np.testing.assert_allclose(
        self.entity.global_vector_to_local_frame(physics, [0, 0, 1]),
        [0, 0, 1], atol=1e-10)

    # 2D vectors; z-component is ignored
    np.testing.assert_allclose(
        self.entity.global_vector_to_local_frame(physics, [0, 1]),
        [1, 0], atol=1e-10)
    np.testing.assert_allclose(
        self.entity.global_vector_to_local_frame(physics, [-1, 0]),
        [0, 1], atol=1e-10)

  def testGlobalMatrixToLocalFrame(self):
    parent = TestEntity()
    parent.mjcf_model.worldbody.add(
        'site', xyaxes=[0, 1, 0, -1, 0, 0]).attach(self.entity.mjcf_model)
    physics = mjcf.Physics.from_mjcf_model(parent.mjcf_model)

    rotation_atob = np.array([[0, 1, 0], [0, 0, -1], [-1, 0, 0]])
    ego_rotation_atob = np.array([[0, 0, -1], [0, -1, 0], [-1, 0, 0]])

    np.testing.assert_allclose(
        self.entity.global_xmat_to_local_frame(physics, rotation_atob),
        ego_rotation_atob, atol=1e-10)

    flat_rotation_atob = np.reshape(rotation_atob, -1)
    flat_rotation_ego_atob = np.reshape(ego_rotation_atob, -1)
    np.testing.assert_allclose(
        self.entity.global_xmat_to_local_frame(
            physics, flat_rotation_atob),
        flat_rotation_ego_atob, atol=1e-10)

  @parameterized.parameters(*_param_product(
      position=[None, [1., 0., -1.]],
      quaternion=[None, _FORTYFIVE_DEGREES_ABOUT_X, _NINETY_DEGREES_ABOUT_Z],
      freejoint=[False, True],
  ))
  def testSetPose(self, position, quaternion, freejoint):
    # Setup entity.
    test_arena = arena.Arena()
    subentity = TestEntity(name='subentity')
    frame = test_arena.attach(subentity)
    if freejoint:
      frame.add('freejoint')

    physics = mjcf.Physics.from_mjcf_model(test_arena.mjcf_model)

    if quaternion is None:
      ground_truth_quat = _NO_ROTATION
    else:
      ground_truth_quat = quaternion

    if position is None:
      ground_truth_pos = np.zeros(shape=(3,))
    else:
      ground_truth_pos = position

    subentity.set_pose(physics, position=position, quaternion=quaternion)

    np.testing.assert_array_equal(physics.bind(frame).xpos, ground_truth_pos)
    np.testing.assert_array_equal(physics.bind(frame).xquat, ground_truth_quat)

  @parameterized.parameters(*_param_product(
      original_position=[[-2, -1, -1.], [1., 0., -1.]],
      position=[None, [1., 0., -1.]],
      original_quaternion=_TEST_ROTATIONS[0],
      quaternion=_TEST_ROTATIONS[1],
      expected_quaternion=_TEST_ROTATIONS[2],
      freejoint=[False, True],
  ))
  def testShiftPose(self, original_position, position, original_quaternion,
                    quaternion, expected_quaternion, freejoint):
    # Setup entity.
    test_arena = arena.Arena()
    subentity = TestEntity(name='subentity')
    frame = test_arena.attach(subentity)
    if freejoint:
      frame.add('freejoint')

    physics = mjcf.Physics.from_mjcf_model(test_arena.mjcf_model)

    # Set the original position
    subentity.set_pose(
        physics, position=original_position, quaternion=original_quaternion)

    if position is None:
      ground_truth_pos = original_position
    else:
      ground_truth_pos = original_position + np.array(position)
    subentity.shift_pose(physics, position=position, quaternion=quaternion)
    np.testing.assert_array_equal(physics.bind(frame).xpos, ground_truth_pos)

    updated_quat = physics.bind(frame).xquat
    np.testing.assert_array_almost_equal(updated_quat, expected_quaternion,
                                         1e-4)

  @parameterized.parameters(False, True)
  def testShiftPoseWithVelocity(self, rotate_velocity):
    # Setup entity.
    test_arena = arena.Arena()
    subentity = TestEntity(name='subentity')
    frame = test_arena.attach(subentity)
    frame.add('freejoint')

    physics = mjcf.Physics.from_mjcf_model(test_arena.mjcf_model)

    # Set the original position
    subentity.set_pose(physics, position=[0., 0., 0.])

    # Set velocity in y dim.
    subentity.set_velocity(physics, [0., 1., 0.])

    # Rotate the entity around the z axis.
    subentity.shift_pose(
        physics, quaternion=[0., 0., 0., 1.], rotate_velocity=rotate_velocity)

    physics.forward()
    updated_position, _ = subentity.get_pose(physics)
    if rotate_velocity:
      # Should not have moved in the y dim.
      np.testing.assert_array_almost_equal(updated_position[1], 0.)
    else:
      # Should not have moved in the x dim.
      np.testing.assert_array_almost_equal(updated_position[0], 0.)


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests for dm_control.locomotion.soccer.soccer_ball."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import absltest
from dm_control import composer
from dm_control import mjcf
from dm_control.entities import props
from dm_control.locomotion.soccer import soccer_ball
from dm_control.locomotion.soccer import team
import numpy as np


class SoccerBallTest(absltest.TestCase):

  def test_detect_hit(self):
    arena = composer.Arena()
    ball = soccer_ball.SoccerBall(radius=0.35, mass=0.045, name='test_ball')
    player = team.Player(
        team=team.Team.HOME,
        walker=props.Primitive(geom_type='sphere', size=(0.1,), name='home'))
    arena.add_free_entity(player.walker)
    ball.register_player(player)
    arena.add_free_entity(ball)

    random_state = np.random.RandomState(42)
    physics = mjcf.Physics.from_mjcf_model(arena.mjcf_model)
    physics.step()

    ball.initialize_episode(physics, random_state)
    ball.before_step(physics, random_state)
    self.assertEqual(ball.hit, False)
    self.assertEqual(ball.repossessed, False)
    self.assertEqual(ball.intercepted, False)
    self.assertIsNone(ball.last_hit)
    self.assertIsNone(ball.dist_between_last_hits)

    ball.after_substep(physics, random_state)
    ball.after_step(physics, random_state)

    self.assertEqual(ball.hit, True)
    self.assertEqual(ball.repossessed, True)
    self.assertEqual(ball.intercepted, True)
    self.assertEqual(ball.last_hit, player)
    # Only one hit registered.
    self.assertIsNone(ball.dist_between_last_hits)

  def test_has_tracking_cameras(self):
    ball = soccer_ball.SoccerBall(radius=0.35, mass=0.045, name='test_ball')
    expected_camera_names = ['ball_cam_near', 'ball_cam', 'ball_cam_far']
    camera_names = [cam.name for cam in ball.mjcf_model.find_all('camera')]
    self.assertCountEqual(expected_camera_names, camera_names)


if __name__ == '__main__':
  absltest.main()



# Copyright 2017-2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests for dm_control.utils.render_executor."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import threading
import time
import unittest

from absl.testing import absltest
from absl.testing import parameterized
from dm_control._render import executor
import mock
from six.moves import range


def enforce_timeout(timeout):
  def wrap(test_func):
    def wrapped_test(self, *args, **kwargs):
      thread = threading.Thread(
          target=test_func, args=((self,) + args), kwargs=kwargs)
      thread.daemon = True
      thread.start()
      thread.join(timeout=timeout)
      self.assertFalse(
          thread.is_alive(),
          msg='Test timed out after {} seconds.'.format(timeout))
    return wrapped_test
  return wrap


class RenderExecutorTest(parameterized.TestCase):

  def _make_executor(self, executor_type):
    if (executor_type == executor.NativeMutexOffloadingRenderExecutor and
        executor_type is None):
      raise unittest.SkipTest(
          'NativeMutexOffloadingRenderExecutor is not available.')
    else:
      return executor_type()

  def test_passthrough_executor_thread(self):
    render_executor = self._make_executor(executor.PassthroughRenderExecutor)
    self.assertIs(render_executor.thread, threading.current_thread())
    render_executor.terminate()

  @parameterized.parameters(executor.OffloadingRenderExecutor,
                            executor.NativeMutexOffloadingRenderExecutor)
  def test_offloading_executor_thread(self, executor_type):
    render_executor = self._make_executor(executor_type)
    self.assertIsNot(render_executor.thread, threading.current_thread())
    render_executor.terminate()

  @parameterized.parameters(executor.PassthroughRenderExecutor,
                            executor.OffloadingRenderExecutor,
                            executor.NativeMutexOffloadingRenderExecutor)
  def test_call_on_correct_thread(self, executor_type):
    render_executor = self._make_executor(executor_type)
    with render_executor.execution_context() as ctx:
      actual_executed_thread = ctx.call(threading.current_thread)
    self.assertIs(actual_executed_thread, render_executor.thread)
    render_executor.terminate()

  @parameterized.parameters(executor.PassthroughRenderExecutor,
                            executor.OffloadingRenderExecutor,
                            executor.NativeMutexOffloadingRenderExecutor)
  def test_multithreaded(self, executor_type):
    render_executor = self._make_executor(executor_type)
    list_length = 5
    shared_list = [None] * list_length

    def fill_list(thread_idx):
      def assign_value(i):
        shared_list[i] = thread_idx
      for _ in range(1000):
        with render_executor.execution_context() as ctx:
          for i in range(list_length):
            ctx.call(assign_value, i)
          # Other threads should be prevented from calling `assign_value` while
          # this thread is inside the `execution_context`.
          self.assertEqual(shared_list, [thread_idx] * list_length)

    threads = [threading.Thread(target=fill_list, args=(i,)) for i in range(9)]
    for thread in threads:
      thread.start()
    for thread in threads:
      thread.join()

    render_executor.terminate()

  @parameterized.parameters(executor.PassthroughRenderExecutor,
                            executor.OffloadingRenderExecutor,
                            executor.NativeMutexOffloadingRenderExecutor)
  def test_exception(self, executor_type):
    render_executor = self._make_executor(executor_type)
    message = 'fake error'
    def raise_value_error():
      raise ValueError(message)
    with render_executor.execution_context() as ctx:
      with self.assertRaisesWithLiteralMatch(ValueError, message):
        ctx.call(raise_value_error)
    render_executor.terminate()

  @parameterized.parameters(executor.PassthroughRenderExecutor,
                            executor.OffloadingRenderExecutor,
                            executor.NativeMutexOffloadingRenderExecutor)
  def test_terminate(self, executor_type):
    render_executor = self._make_executor(executor_type)
    cleanup = mock.MagicMock()
    render_executor.terminate(cleanup)
    cleanup.assert_called_once_with()

  @parameterized.parameters(executor.PassthroughRenderExecutor,
                            executor.OffloadingRenderExecutor,
                            executor.NativeMutexOffloadingRenderExecutor)
  def test_call_outside_of_context(self, executor_type):
    render_executor = self._make_executor(executor_type)
    func = mock.MagicMock()
    with self.assertRaisesWithLiteralMatch(
        RuntimeError, executor.render_executor._NOT_IN_CONTEXT):
      render_executor.call(func)
    # Also test that the locked flag is properly cleared when leaving a context.
    with render_executor.execution_context():
      render_executor.call(lambda: None)
    with self.assertRaisesWithLiteralMatch(
        RuntimeError, executor.render_executor._NOT_IN_CONTEXT):
      render_executor.call(func)
    func.assert_not_called()
    render_executor.terminate()

  @parameterized.parameters(executor.PassthroughRenderExecutor,
                            executor.OffloadingRenderExecutor,
                            executor.NativeMutexOffloadingRenderExecutor)
  def test_call_after_terminate(self, executor_type):
    render_executor = self._make_executor(executor_type)
    render_executor.terminate()
    func = mock.MagicMock()
    with self.assertRaisesWithLiteralMatch(
        RuntimeError, executor.render_executor._ALREADY_TERMINATED):
      with render_executor.execution_context() as ctx:
        ctx.call(func)
    func.assert_not_called()

  @parameterized.parameters(executor.PassthroughRenderExecutor,
                            executor.OffloadingRenderExecutor,
                            executor.NativeMutexOffloadingRenderExecutor)
  def test_locking(self, executor_type):
    render_executor = self._make_executor(executor_type)
    other_thread_context_entered = threading.Condition()
    other_thread_context_done = [False]
    def other_thread_func():
      with render_executor.execution_context():
        with other_thread_context_entered:
          other_thread_context_entered.notify()
        time.sleep(1)
        other_thread_context_done[0] = True
    other_thread = threading.Thread(target=other_thread_func)
    with other_thread_context_entered:
      other_thread.start()
      other_thread_context_entered.wait()
    with render_executor.execution_context():
      self.assertTrue(
          other_thread_context_done[0],
          msg=('Main thread should not be able to enter the execution context '
               'until the other thread is done.'))

  @parameterized.parameters(executor.PassthroughRenderExecutor,
                            executor.OffloadingRenderExecutor,
                            executor.NativeMutexOffloadingRenderExecutor)
  @enforce_timeout(timeout=5.)
  def test_reentrant_locking(self, executor_type):
    render_executor = self._make_executor(executor_type)
    def triple_lock(render_executor):
      with render_executor.execution_context():
        with render_executor.execution_context():
          with render_executor.execution_context():
            pass
    triple_lock(render_executor)

  @parameterized.parameters(executor.PassthroughRenderExecutor,
                            executor.OffloadingRenderExecutor,
                            executor.NativeMutexOffloadingRenderExecutor)
  @enforce_timeout(timeout=5.)
  def test_no_deadlock_in_callbacks(self, executor_type):
    render_executor = self._make_executor(executor_type)
    # This test times out in the event of a deadlock.
    def callback():
      with render_executor.execution_context() as ctx:
        ctx.call(lambda: None)
    with render_executor.execution_context() as ctx:
      ctx.call(callback)

if __name__ == '__main__':
  absltest.main()



# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Maze-based arenas."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import string

from absl import logging
from dm_control import composer
from dm_control.composer.observation import observable
from dm_control.locomotion.arenas import assets as locomotion_arenas_assets
from dm_control.locomotion.arenas import covering
import labmaze
import numpy as np
import six
from six.moves import range
from six.moves import zip


# Put all "actual" wall geoms in a separate group since they are not rendered.
_WALL_GEOM_GROUP = 3

_TOP_CAMERA_DISTANCE = 100
_TOP_CAMERA_Y_PADDING_FACTOR = 1.1

_DEFAULT_WALL_CHAR = '*'
_DEFAULT_FLOOR_CHAR = '.'


class MazeWithTargets(composer.Arena):
  """A 2D maze with target positions specified by a LabMaze-style text maze."""

  def _build(self, maze, xy_scale=2.0, z_height=2.0,
             skybox_texture=None, wall_textures=None, floor_textures=None,
             aesthetic='default', name='maze'):
    """Initializes this maze arena.

    Args:
      maze: A `labmaze.BaseMaze` instance.
      xy_scale: The size of each maze cell in metres.
      z_height: The z-height of the maze in metres.
      skybox_texture: (optional) A `composer.Entity` that provides a texture
        asset for the skybox.
      wall_textures: (optional) Either a `composer.Entity` that provides texture
        assets for the maze walls, or a dict mapping printable characters to
        such Entities. In the former case, the maze walls are assumed to be
        represented by '*' in the maze's entity layer. In the latter case,
        the dict's keys specify the different characters that can be present
        in the maze's entity layer, and the dict's values are the corresponding
        texture providers.
      floor_textures: (optional) A `composer.Entity` that provides texture
        assets for the maze floor. Unlike with walls, we do not currently
        support per-variation floor texture. Instead, we sample textures from
        the same texture provider for each variation in the variations layer.
      aesthetic: option to adjust the material properties and skybox
      name: (optional) A string, the name of this arena.
    """
    super(MazeWithTargets, self)._build(name)
    self._maze = maze
    self._xy_scale = xy_scale
    self._z_height = z_height

    self._x_offset = (self._maze.width - 1) / 2
    self._y_offset = (self._maze.height - 1) / 2

    self._mjcf_root.default.geom.rgba = [1, 1, 1, 1]

    if aesthetic != 'default':
      sky_info = locomotion_arenas_assets.get_sky_texture_info(aesthetic)
      texturedir = locomotion_arenas_assets.get_texturedir(aesthetic)
      self._mjcf_root.compiler.texturedir = texturedir
      self._skybox = self._mjcf_root.asset.add(
          'texture', name='aesthetic_skybox', file=sky_info.file,
          type='skybox', gridsize=sky_info.gridsize,
          gridlayout=sky_info.gridlayout)
    elif skybox_texture:
      self._skybox_texture = skybox_texture.texture
      self.attach(skybox_texture)
    else:
      self._skybox_texture = self._mjcf_root.asset.add(
          'texture', type='skybox', name='skybox', builtin='gradient',
          rgb1=[.4, .6, .8], rgb2=[0, 0, 0], width=100, height=100)

    self._texturing_geom_names = []
    self._texturing_material_names = []
    if wall_textures:
      if isinstance(wall_textures, dict):
        for texture_provider in set(wall_textures.values()):
          self.attach(texture_provider)
        self._wall_textures = {
            wall_char: texture_provider.textures
            for wall_char, texture_provider in six.iteritems(wall_textures)
        }
      else:
        self.attach(wall_textures)
        self._wall_textures = {_DEFAULT_WALL_CHAR: wall_textures.textures}
    else:
      self._wall_textures = {_DEFAULT_WALL_CHAR: [self._mjcf_root.asset.add(
          'texture', type='2d', name='wall', builtin='flat',
          rgb1=[.8, .8, .8], width=100, height=100)]}

    if aesthetic != 'default':
      ground_info = locomotion_arenas_assets.get_ground_texture_info(aesthetic)
      self._floor_textures = [
          self._mjcf_root.asset.add(
              'texture',
              name='aesthetic_texture_main',
              file=ground_info.file,
              type=ground_info.type),
          self._mjcf_root.asset.add(
              'texture',
              name='aesthetic_texture',
              file=ground_info.file,
              type=ground_info.type)
      ]
    elif floor_textures:
      self._floor_textures = floor_textures.textures
      self.attach(floor_textures)
    else:
      self._floor_textures = [self._mjcf_root.asset.add(
          'texture', type='2d', name='floor', builtin='flat',
          rgb1=[.2, .2, .2], width=100, height=100)]

    ground_x = ((self._maze.width - 1) + 1) * (xy_scale / 2)
    ground_y = ((self._maze.height - 1) + 1) * (xy_scale / 2)
    self._mjcf_root.worldbody.add(
        'geom', name='ground', type='plane',
        pos=[0, 0, 0], size=[ground_x, ground_y, 1], rgba=[0, 0, 0, 0])

    self._maze_body = self._mjcf_root.worldbody.add('body', name='maze_body')

    self._mjcf_root.visual.map.znear = 0.0005

    # Choose the FOV so that the maze always fits nicely within the frame
    # irrespective of actual maze size.
    maze_size = max(self._maze.width, self._maze.height)
    top_camera_fovy = (360 / np.pi) * np.arctan2(
        _TOP_CAMERA_Y_PADDING_FACTOR * maze_size * self._xy_scale / 2,
        _TOP_CAMERA_DISTANCE)
    self._top_camera = self._mjcf_root.worldbody.add(
        'camera', name='top_camera',
        pos=[0, 0, _TOP_CAMERA_DISTANCE], zaxis=[0, 0, 1], fovy=top_camera_fovy)

    self._target_positions = ()
    self._spawn_positions = ()

    self._text_maze_regenerated_hook = None
    self._tile_geom_names = {}

  def _build_observables(self):
    return MazeObservables(self)

  @property
  def top_camera(self):
    return self._top_camera

  @property
  def xy_scale(self):
    return self._xy_scale

  @property
  def z_height(self):
    return self._z_height

  @property
  def maze(self):
    return self._maze

  @property
  def text_maze_regenerated_hook(self):
    """A callback that is executed after the LabMaze object is regenerated."""
    return self._text_maze_modifier

  @text_maze_regenerated_hook.setter
  def text_maze_regenerated_hook(self, hook):
    self._text_maze_regenerated_hook = hook

  @property
  def target_positions(self):
    """A tuple of Cartesian target positions generated for the current maze."""
    return self._target_positions

  @property
  def spawn_positions(self):
    """The Cartesian position at which the agent should be spawned."""
    return self._spawn_positions

  @property
  def target_grid_positions(self):
    """A tuple of grid coordinates of targets generated for the current maze."""
    return self._target_grid_positions

  @property
  def spawn_grid_positions(self):
    """The grid-coordinate position at which the agent should be spawned."""
    return self._spawn_grid_positions

  def regenerate(self, random_state=np.random.RandomState()):
    """Generates a new maze layout."""
    del random_state
    self._maze.regenerate()
    logging.debug('GENERATED MAZE:\n%s', self._maze.entity_layer)
    self._find_spawn_and_target_positions()

    if self._text_maze_regenerated_hook:
      self._text_maze_regenerated_hook()

    # Remove old texturing planes.
    for geom_name in self._texturing_geom_names:
      del self._mjcf_root.worldbody.geom[geom_name]
    self._texturing_geom_names = []

    # Remove old texturing materials.
    for material_name in self._texturing_material_names:
      del self._mjcf_root.asset.material[material_name]
    self._texturing_material_names = []

    # Remove old actual-wall geoms.
    self._maze_body.geom.clear()

    self._current_wall_texture = {
        wall_char: np.random.choice(wall_textures)
        for wall_char, wall_textures in six.iteritems(self._wall_textures)
    }

    for wall_char in self._wall_textures:
      self._make_wall_geoms(wall_char)
    self._make_floor_variations()

  def _make_wall_geoms(self, wall_char):
    walls = covering.make_walls(
        self._maze.entity_layer, wall_char=wall_char, make_odd_sized_walls=True)
    for i, wall in enumerate(walls):
      wall_mid = covering.GridCoordinates(
          (wall.start.y + wall.end.y - 1) / 2,
          (wall.start.x + wall.end.x - 1) / 2)
      wall_pos = np.array([(wall_mid.x - self._x_offset) * self._xy_scale,
                           -(wall_mid.y - self._y_offset) * self._xy_scale,
                           self._z_height / 2])
      wall_size = np.array([(wall.end.x - wall_mid.x - 0.5) * self._xy_scale,
                            (wall.end.y - wall_mid.y - 0.5) * self._xy_scale,
                            self._z_height / 2])
      self._maze_body.add('geom', name='wall{}_{}'.format(wall_char, i),
                          type='box', pos=wall_pos, size=wall_size,
                          group=_WALL_GEOM_GROUP)
      self._make_wall_texturing_planes(wall_char, i, wall_pos, wall_size)

  def _make_wall_texturing_planes(self, wall_char, wall_id,
                                  wall_pos, wall_size):
    xyaxes = {
        'x': {-1: [0, -1, 0, 0, 0, 1], 1: [0, 1, 0, 0, 0, 1]},
        'y': {-1: [1, 0, 0, 0, 0, 1], 1: [-1, 0, 0, 0, 0, 1]},
        'z': {-1: [-1, 0, 0, 0, 1, 0], 1: [1, 0, 0, 0, 1, 0]}
    }
    for direction_index, direction in enumerate(('x', 'y', 'z')):
      index = list(i for i in range(3) if i != direction_index)
      delta_vector = np.array([int(i == direction_index) for i in range(3)])
      material_name = 'wall{}_{}_{}'.format(wall_char, wall_id, direction)
      self._texturing_material_names.append(material_name)
      mat = self._mjcf_root.asset.add(
          'material', name=material_name,
          texture=self._current_wall_texture[wall_char],
          texrepeat=(2 * wall_size[index] / self._xy_scale))
      for sign, sign_name in zip((-1, 1), ('neg', 'pos')):
        if direction == 'z' and sign == -1:
          continue
        geom_name = (
            'wall{}_{}_texturing_{}_{}'.format(
                wall_char, wall_id, sign_name, direction))
        self._texturing_geom_names.append(geom_name)
        self._mjcf_root.worldbody.add(
            'geom', type='plane', name=geom_name,
            pos=(wall_pos + sign * delta_vector * wall_size),
            size=np.concatenate([wall_size[index], [self._xy_scale]]),
            xyaxes=xyaxes[direction][sign], material=mat,
            contype=0, conaffinity=0)

  def _make_floor_variations(self, build_tile_geoms_fn=None):
    """Builds the floor tiles.

    Args:
      build_tile_geoms_fn: An optional callable returning floor tile geoms.
        If not passed, the floor will be built using a default covering method.
        Takes a kwarg `wall_char` that can be used control how active floor
        tiles are selected.
    """
    main_floor_texture = np.random.choice(self._floor_textures)
    for variation in _DEFAULT_FLOOR_CHAR + string.ascii_uppercase:
      if variation not in self._maze.variations_layer:
        break

      if build_tile_geoms_fn is None:
        # Break the floor variation down to odd-sized tiles.
        tiles = covering.make_walls(self._maze.variations_layer,
                                    wall_char=variation,
                                    make_odd_sized_walls=True)
      else:
        tiles = build_tile_geoms_fn(wall_char=variation)

      # Sample a texture that's not the same as the main floor texture.
      variation_texture = main_floor_texture
      if variation != _DEFAULT_FLOOR_CHAR:
        if len(self._floor_textures) == 1:
          return
        else:
          while variation_texture is main_floor_texture:
            variation_texture = np.random.choice(self._floor_textures)

      for i, tile in enumerate(tiles):
        tile_mid = covering.GridCoordinates(
            (tile.start.y + tile.end.y - 1) / 2,
            (tile.start.x + tile.end.x - 1) / 2)
        tile_pos = np.array([(tile_mid.x - self._x_offset) * self._xy_scale,
                             -(tile_mid.y - self._y_offset) * self._xy_scale,
                             0.0])
        tile_size = np.array([(tile.end.x - tile_mid.x - 0.5) * self._xy_scale,
                              (tile.end.y - tile_mid.y - 0.5) * self._xy_scale,
                              self._xy_scale])
        if variation == _DEFAULT_FLOOR_CHAR:
          tile_name = 'floor_{}'.format(i)
        else:
          tile_name = 'floor_{}_{}'.format(variation, i)
        self._tile_geom_names[tile.start] = tile_name
        self._texturing_material_names.append(tile_name)
        self._texturing_geom_names.append(tile_name)
        material = self._mjcf_root.asset.add(
            'material', name=tile_name, texture=variation_texture,
            texrepeat=(2 * tile_size[[0, 1]] / self._xy_scale))
        self._mjcf_root.worldbody.add(
            'geom', name=tile_name, type='plane', material=material,
            pos=tile_pos, size=tile_size, contype=0, conaffinity=0)

  @property
  def ground_geoms(self):
    return tuple([
        geom for geom in self.mjcf_model.find_all('geom')
        if 'ground' in geom.name
    ])

  def find_token_grid_positions(self, tokens):
    out = {token: [] for token in tokens}
    for y in range(self._maze.entity_layer.shape[0]):
      for x in range(self._maze.entity_layer.shape[1]):
        for token in tokens:
          if self._maze.entity_layer[y, x] == token:
            out[token].append((y, x))
    return out

  def grid_to_world_positions(self, grid_positions):
    out = []
    for y, x in grid_positions:
      out.append(np.array([(x - self._x_offset) * self._xy_scale,
                           -(y - self._y_offset) * self._xy_scale,
                           0.0]))
    return out

  def world_to_grid_positions(self, world_positions):
    out = []
    # the order of x, y is reverse between grid positions format and
    # world positions format.
    for x, y, _ in world_positions:
      out.append(np.array([self._y_offset - y / self._xy_scale,
                           self._x_offset + x / self._xy_scale]))
    return out

  def _find_spawn_and_target_positions(self):
    grid_positions = self.find_token_grid_positions([
        labmaze.defaults.OBJECT_TOKEN, labmaze.defaults.SPAWN_TOKEN])
    self._target_grid_positions = tuple(
        grid_positions[labmaze.defaults.OBJECT_TOKEN])
    self._spawn_grid_positions = tuple(
        grid_positions[labmaze.defaults.SPAWN_TOKEN])
    self._target_positions = tuple(
        self.grid_to_world_positions(self._target_grid_positions))
    self._spawn_positions = tuple(
        self.grid_to_world_positions(self._spawn_grid_positions))


class MazeObservables(composer.Observables):

  @composer.observable
  def top_camera(self):
    return observable.MJCFCamera(self._entity.top_camera)


class RandomMazeWithTargets(MazeWithTargets):
  """A randomly generated 2D maze with target positions."""

  def _build(self,
             x_cells,
             y_cells,
             xy_scale=2.0,
             z_height=2.0,
             max_rooms=labmaze.defaults.MAX_ROOMS,
             room_min_size=labmaze.defaults.ROOM_MIN_SIZE,
             room_max_size=labmaze.defaults.ROOM_MAX_SIZE,
             spawns_per_room=labmaze.defaults.SPAWN_COUNT,
             targets_per_room=labmaze.defaults.OBJECT_COUNT,
             max_variations=labmaze.defaults.MAX_VARIATIONS,
             simplify=labmaze.defaults.SIMPLIFY,
             skybox_texture=None,
             wall_textures=None,
             floor_textures=None,
             aesthetic='default',
             name='random_maze'):
    """Initializes this random maze arena.

    Args:
      x_cells: The number of cells along the x-direction of the maze. Must be
        an odd integer.
      y_cells: The number of cells along the y-direction of the maze. Must be
        an odd integer.
      xy_scale: The size of each maze cell in metres.
      z_height: The z-height of the maze in metres.
      max_rooms: (optional) The maximum number of rooms in each generated maze.
      room_min_size: (optional) The minimum size of each room generated.
      room_max_size: (optional) The maximum size of each room generated.
      spawns_per_room: (optional) Number of spawn points
        to generate in each room.
      targets_per_room: (optional) Number of targets to generate in each room.
      max_variations: (optional) Maximum number of variations to generate
        in the variations layer.
      simplify: (optional) flag to simplify the maze.
      skybox_texture: (optional) A `composer.Entity` that provides a texture
        asset for the skybox.
      wall_textures: (optional) A `composer.Entity` that provides texture
        assets for the maze walls.
      floor_textures: (optional) A `composer.Entity` that provides texture
        assets for the maze floor.
      aesthetic: option to adjust the material properties and skybox
      name: (optional) A string, the name of this arena.
    """
    random_seed = np.random.randint(2147483648)  # 2**31
    super(RandomMazeWithTargets, self)._build(
        maze=labmaze.RandomMaze(
            height=y_cells,
            width=x_cells,
            max_rooms=max_rooms,
            room_min_size=room_min_size,
            room_max_size=room_max_size,
            max_variations=max_variations,
            spawns_per_room=spawns_per_room,
            objects_per_room=targets_per_room,
            simplify=simplify,
            random_seed=random_seed),
        xy_scale=xy_scale,
        z_height=z_height,
        skybox_texture=skybox_texture,
        wall_textures=wall_textures,
        floor_textures=floor_textures,
        aesthetic=aesthetic,
        name=name)



# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tools for adding custom cameras to the arena."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

from dm_control.composer.observation import observable


CameraSpec = collections.namedtuple('CameraSpec', ['name', 'pos', 'xyaxes'])

# Custom cameras that may be added to the arena for particular tasks.
FRONT_CLOSE = CameraSpec(
    name='front_close',
    pos=(0., -0.6, 0.75),
    xyaxes=(1., 0., 0., 0., 0.7, 0.75)
)

FRONT_FAR = CameraSpec(
    name='front_far',
    pos=(0., -0.8, 1.),
    xyaxes=(1., 0., 0., 0., 0.7, 0.75)
)

TOP_DOWN = CameraSpec(
    name='top_down',
    pos=(0., 0., 2.5),
    xyaxes=(1., 0., 0., 0., 1., 0.)
)

LEFT_CLOSE = CameraSpec(
    name='left_close',
    pos=(-0.6, 0., 0.75),
    xyaxes=(0., -1., 0., 0.7, 0., 0.75)
)

RIGHT_CLOSE = CameraSpec(
    name='right_close',
    pos=(0.6, 0., 0.75),
    xyaxes=(0., 1., 0., -0.7, 0., 0.75)
)


def add_camera_observables(entity, obs_settings, *camera_specs):
  """Adds cameras to an entity's worldbody and configures observables for them.

  Args:
    entity: A `composer.Entity`.
    obs_settings: An `observations.ObservationSettings` instance.
    *camera_specs: Instances of `CameraSpec`.

  Returns:
    A `collections.OrderedDict` keyed on camera names, containing pre-configured
    `observable.MJCFCamera` instances.
  """
  obs_dict = collections.OrderedDict()
  for spec in camera_specs:
    camera = entity.mjcf_model.worldbody.add('camera', **spec._asdict())
    obs = observable.MJCFCamera(camera)
    obs.configure(**obs_settings.camera._asdict())
    obs_dict[spec.name] = obs
  return obs_dict



# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests for dm_control.locomotion.walkers.base."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import absltest
from dm_control import mjcf
from dm_control.locomotion.walkers import base

import numpy as np


class FakeWalker(base.Walker):

  def _build(self):
    self._mjcf_root = mjcf.RootElement(model='walker')
    self._torso_body = self._mjcf_root.worldbody.add(
        'body', name='torso', xyaxes=[0, 1, 0, -1, 0, 0])

  @property
  def mjcf_model(self):
    return self._mjcf_root

  @property
  def actuators(self):
    return []

  @property
  def root_body(self):
    return self._torso_body

  @property
  def observable_joints(self):
    return []


class BaseWalkerTest(absltest.TestCase):

  def testTransformVectorToEgocentricFrame(self):
    walker = FakeWalker()
    physics = mjcf.Physics.from_mjcf_model(walker.mjcf_model)

    # 3D vectors
    np.testing.assert_allclose(
        walker.transform_vec_to_egocentric_frame(physics, [0, 1, 0]), [1, 0, 0],
        atol=1e-10)
    np.testing.assert_allclose(
        walker.transform_vec_to_egocentric_frame(physics, [-1, 0, 0]),
        [0, 1, 0],
        atol=1e-10)
    np.testing.assert_allclose(
        walker.transform_vec_to_egocentric_frame(physics, [0, 0, 1]), [0, 0, 1],
        atol=1e-10)

    # 2D vectors; z-component is ignored
    np.testing.assert_allclose(
        walker.transform_vec_to_egocentric_frame(physics, [0, 1]), [1, 0],
        atol=1e-10)
    np.testing.assert_allclose(
        walker.transform_vec_to_egocentric_frame(physics, [-1, 0]), [0, 1],
        atol=1e-10)

  def testTransformMatrixToEgocentricFrame(self):
    walker = FakeWalker()
    physics = mjcf.Physics.from_mjcf_model(walker.mjcf_model)

    rotation_atob = np.array([[0, 1, 0], [0, 0, -1], [-1, 0, 0]])
    ego_rotation_atob = np.array([[0, 0, -1], [0, -1, 0], [-1, 0, 0]])

    np.testing.assert_allclose(
        walker.transform_xmat_to_egocentric_frame(physics, rotation_atob),
        ego_rotation_atob, atol=1e-10)

    flat_rotation_atob = np.reshape(rotation_atob, -1)
    flat_rotation_ego_atob = np.reshape(ego_rotation_atob, -1)
    np.testing.assert_allclose(
        walker.transform_xmat_to_egocentric_frame(physics, flat_rotation_atob),
        flat_rotation_ego_atob, atol=1e-10)


if __name__ == '__main__':
  absltest.main()



# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Classes to represent MJCF elements in the object model."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import copy
import os
import sys

from dm_control.mjcf import attribute as attribute_types
from dm_control.mjcf import base
from dm_control.mjcf import constants
from dm_control.mjcf import copier
from dm_control.mjcf import debugging
from dm_control.mjcf import namescope
from dm_control.mjcf import schema
from dm_control.mujoco.wrapper import util
from lxml import etree
import numpy as np
import six
from six.moves import zip


_raw_property = property  # pylint: disable=invalid-name


_CONFLICT_BEHAVIOR_FUNC = {'min': min, 'max': max}


def property(method):  # pylint: disable=redefined-builtin
  """Modifies `@property` to keep track of any `AttributeError` raised.

  Our `Element` implementations overrides the `__getattr__` method. This does
  not interact well with `@property`: if a `property`'s code is buggy so as to
  raise an `AttributeError`, then Python would silently discard it and redirect
  to our `__getattr__` instead, leading to an uninformative stack trace. This
  makes it very difficult to debug issues that involve properties.

  To remedy this, we modify `@property` within this module to store any
  `AttributeError` raised within the respective `Element` object. Then, in our
  `__getattr__` logic, we could re-raise it to preserve the original stack
  trace.

  The reason that this is not implemented as a different decorator is that we
  could accidentally use @property on a new method. This would work fine until
  someone triggers a subtle bug. This is when a proper trace would be most
  useful, but we would still end up with a strange undebuggable stack trace
  anyway.

  Note that at the end of this module, we have a `del property` to prevent this
  override from being broadcasted externally.

  Args:
    method: The method that is being decorated.

  Returns:
    A `property` corresponding to the decorated method.
  """
  def _mjcf_property(self):
    try:
      return method(self)
    except:  # pylint: disable=bare-except
      err_type, err, tb = sys.exc_info()
      tb_next = tb.tb_next
      if isinstance(err, AttributeError):
        self._last_attribute_error = err_type, err, tb_next  # pylint: disable=protected-access
      six.reraise(err_type, err, tb_next)
  return _raw_property(_mjcf_property)


def _make_element(spec, parent, attributes=None):
  """Helper function to generate the right kind of Element given a spec."""
  if (spec.name == constants.WORLDBODY
      or (spec.name == constants.SITE
          and (parent.tag == constants.BODY
               or parent.tag == constants.WORLDBODY))):
    return _AttachableElement(spec, parent, attributes)
  elif isinstance(parent, _AttachmentFrame):
    return _AttachmentFrameChild(spec, parent, attributes)
  elif spec.name == constants.DEFAULT:
    return _DefaultElement(spec, parent, attributes)
  elif spec.name == constants.ACTUATOR:
    return _ActuatorElement(spec, parent, attributes)
  else:
    return _ElementImpl(spec, parent, attributes)


_DEFAULT_NAME_FROM_FILENAME = frozenset(['mesh', 'hfield', 'texture'])


class _ElementImpl(base.Element):
  """Actual implementation of a generic MJCF element object."""
  __slots__ = ['__weakref__', '_spec', '_parent', '_attributes', '_children',
               '_own_attributes', '_attachments', '_is_removed', '_init_stack',
               '_is_worldbody', '_cached_namescope', '_cached_root',
               '_cached_full_identifier', '_cached_revision',
               '_last_attribute_error']

  def __init__(self, spec, parent, attributes=None):
    attributes = attributes or {}

    # For certain `asset` elements the `name` attribute can be omitted, in which
    # case the name will be the filename without the leading path and extension.
    # See http://www.mujoco.org/book/XMLreference.html#asset.
    if ('name' not in attributes
        and 'file' in attributes
        and spec.name in _DEFAULT_NAME_FROM_FILENAME):
      _, filename = os.path.split(attributes['file'])
      basename, _ = os.path.splitext(filename)
      attributes['name'] = basename

    self._spec = spec
    self._parent = parent
    self._attributes = collections.OrderedDict()
    self._own_attributes = None
    self._children = []
    self._attachments = collections.OrderedDict()
    self._is_removed = False
    self._is_worldbody = (self.tag == 'worldbody')

    if self._parent:
      self._cached_namescope = self._parent.namescope
      self._cached_root = self._parent.root
    self._cached_full_identifier = ''
    self._cached_revision = -1

    self._last_attribute_error = None

    if debugging.debug_mode():
      self._init_stack = debugging.get_current_stack_trace()

    with debugging.freeze_current_stack_trace():
      for child_spec in six.itervalues(self._spec.children):
        if not (child_spec.repeated or child_spec.on_demand):
          self._children.append(_make_element(spec=child_spec, parent=self))

      if constants.DCLASS in attributes:
        attributes[constants.CLASS] = attributes[constants.DCLASS]
        del attributes[constants.DCLASS]

      for attribute_name in six.iterkeys(attributes):
        self._check_valid_attribute(attribute_name)

      for attribute_spec in six.itervalues(self._spec.attributes):
        value = None
        # Some Reference attributes refer to a namespace that is specified
        # via another attribute. We therefore have to set things up for
        # the additional indirection.
        if attribute_spec.type is attribute_types.Reference:
          reference_namespace = (
              attribute_spec.other_kwargs['reference_namespace'])
          if reference_namespace.startswith(
              constants.INDIRECT_REFERENCE_NAMESPACE_PREFIX):
            attribute_spec = copy.deepcopy(attribute_spec)
            namespace_attrib_name = reference_namespace[
                len(constants.INDIRECT_REFERENCE_NAMESPACE_PREFIX):]
            attribute_spec.other_kwargs['reference_namespace'] = (
                self._attributes[namespace_attrib_name])
        if attribute_spec.name in attributes:
          value = attributes[attribute_spec.name]
        try:
          self._attributes[attribute_spec.name] = attribute_spec.type(
              name=attribute_spec.name,
              required=attribute_spec.required,
              conflict_allowed=attribute_spec.conflict_allowed,
              conflict_behavior=attribute_spec.conflict_behavior,
              parent=self, value=value, **attribute_spec.other_kwargs)
        except:  # pylint: disable=bare-except
          # On failure, clear attributes already created
          for attribute_obj in six.itervalues(self._attributes):
            attribute_obj._force_clear()  # pylint: disable=protected-access
          # Then raise a meaningful error
          err_type, err, traceback = sys.exc_info()
          message = (
              'during initialization of attribute {!r} of element <{}>: {}'
              .format(attribute_spec.name, self._spec.name, err))
          six.reraise(err_type, err_type(message), traceback)

  def get_init_stack(self):
    """Gets the stack trace where this element was first initialized."""
    if debugging.debug_mode():
      return self._init_stack

  def get_last_modified_stacks_for_all_attributes(self):
    """Gets a dict of stack traces where each attribute was last modified."""
    return collections.OrderedDict(
        [(name, self._attributes[name].last_modified_stack)
         for name in six.iterkeys(self._spec.attributes)])

  def is_same_as(self, other):
    """Checks whether another element is semantically equivalent to this one.

    Two elements are considered equivalent if they have the same
    specification (i.e. same tag appearing in the same context), the same
    attribute values, and all of their children are equivalent. The ordering
    of non-repeated children is not important for this comparison, while
    the ordering of repeated children are important only amongst the same
    type* of children. In other words, for two bodies to be considered
    equivalent, their child sites must appear in the same order, and their
    child geoms must appear in the same order, but permutations between sites
    and geoms are disregarded. (The only exception is in tendon definition,
    where strict ordering of all children is necessary for equivalence.)

    *Note that the notion of "same type" in this function is very loose:
    for example different actuator element subtypes are treated as separate
    types when children ordering is considered. Therefore, two <actuator>
    elements might be considered equivalent even though they result in different
    orderings of `mjData.ctrl` when compiled. As it stands, this function
    is designed primarily as a testing aid and should not be used to guarantee
    that models are actually identical.

    Args:
      other: An `mjcf.Element`

    Returns:
      `True` if `other` element is semantically equivalent to this one.
    """
    if other is None or other.spec != self._spec:
      return False

    for attribute_name in six.iterkeys(self._spec.attributes):
      attribute = self._attributes[attribute_name]
      other_attribute = getattr(other, attribute_name)
      if isinstance(attribute.value, base.Element):
        if attribute.value.full_identifier != other_attribute.full_identifier:
          return False
      elif not np.all(attribute.value == other_attribute):
        return False

    if (self._parent and
        self._parent.tag == constants.TENDON and
        self._parent.parent == self.root):
      return self._tendon_has_same_children_as(other)
    else:
      return self._has_same_children_as(other)

  def _has_same_children_as(self, other):
    """Helper function to check whether another element has the same children.

    See docstring for `is_same_as` for explanation about the treatment of
    children ordering.

    Args:
      other: An `mjcf.Element`

    Returns:
      A boolean
    """
    for child_name, child_spec in six.iteritems(self._spec.children):
      child = self.get_children(child_name)
      other_child = getattr(other, child_name)
      if not child_spec.repeated:
        if ((child is None and other_child is not None) or
            (child is not None and not child.is_same_as(other_child))):
          return False
      else:
        if len(child) != len(other_child):
          return False
        else:
          for grandchild, other_grandchild in zip(child, other_child):
            if not grandchild.is_same_as(other_grandchild):
              return False
    return True

  def _tendon_has_same_children_as(self, other):
    return all(child.is_same_as(other_child)
               for child, other_child
               in zip(self.all_children(), other.all_children()))

  def _alias_attributes_dict(self, other):
    if self._own_attributes is None:
      self._own_attributes = self._attributes
    self._attributes = other

  def _restore_attributes_dict(self):
    if self._own_attributes is not None:
      for attribute_name, attribute in six.iteritems(self._attributes):
        self._own_attributes[attribute_name].value = attribute.value
      self._attributes = self._own_attributes
      self._own_attributes = None

  @property
  def tag(self):
    return self._spec.name

  @property
  def spec(self):
    return self._spec

  @property
  def parent(self):
    return self._parent

  @property
  def namescope(self):
    return self._cached_namescope

  @property
  def root(self):
    return self._cached_root

  def prefixed_identifier(self, prefix_root):
    if not self._spec.identifier and not self._is_worldbody:
      return None
    elif self._is_worldbody:
      prefix = self.namescope.full_prefix(prefix_root=prefix_root)
      return prefix or 'world'
    else:
      full_identifier = (
          self._attributes[self._spec.identifier].to_xml_string(
              prefix_root=prefix_root))
      if full_identifier:
        return full_identifier
      else:
        prefix = self.namescope.full_prefix(prefix_root=prefix_root)
        prefix = prefix or constants.PREFIX_SEPARATOR
        return prefix + self._default_identifier

  @property
  def full_identifier(self):
    """Fully-qualified identifier used for this element in the generated XML."""
    if self.namescope.revision > self._cached_revision:
      self._cached_full_identifier = self.prefixed_identifier(
          prefix_root=self.namescope.root)
      self._cached_revision = self.namescope.revision
    return self._cached_full_identifier

  @property
  def _default_identifier(self):
    """The default identifier used if this element is not named by the user."""
    if not self._spec.identifier:
      return None
    else:
      siblings = self.root.find_all(self._spec.namespace,
                                    exclude_attachments=True)
      return '{separator}unnamed_{namespace}_{index}'.format(
          separator=constants.PREFIX_SEPARATOR,
          namespace=self._spec.namespace,
          index=siblings.index(self))

  def __dir__(self):
    out_dir = set()
    classes = (type(self),)
    while classes:
      super_classes = set()
      for klass in classes:
        out_dir.update(klass.__dict__)
        super_classes.update(klass.__bases__)
      classes = super_classes
    out_dir.update(self._spec.children)
    out_dir.update(self._spec.attributes)
    if constants.CLASS in out_dir:
      out_dir.remove(constants.CLASS)
      out_dir.add(constants.DCLASS)
    return sorted(out_dir)

  def find(self, namespace, identifier):
    """Finds an element with a particular identifier.

    This function allows the direct access to an arbitrarily deeply nested
    child element by name, without the need to manually traverse through the
    object tree. The `namespace` argument specifies the kind of element to
    find. In most cases, this corresponds to the element's XML tag name.
    However, if an element has multiple specialized tags, then the namespace
    corresponds to the tag name of the most general element of that kind.
    For example, `namespace='joint'` would search for `<joint>` and
    `<freejoint>`, while `namespace='actuator'` would search for `<general>`,
    `<motor>`, `<position>`, `<velocity>`, and `<cylinder>`.

    Args:
      namespace: A string specifying the namespace being searched. See the
        docstring above for explanation.
      identifier: The identifier string of the desired element.

    Returns:
      An `mjcf.Element` object, or `None` if an element with the specified
      identifier is not found.

    Raises:
      ValueError: if either `namespace` or `identifier` is not a string, or if
        `namespace` is not a valid namespace.
    """
    if not isinstance(namespace, six.string_types):
      raise ValueError(
          '`namespace` should be a string: got {!r}'.format(namespace))
    if not isinstance(identifier, six.string_types):
      raise ValueError(
          '`identifier` should be a string: got {!r}'.format(identifier))
    if namespace not in schema.FINDABLE_NAMESPACES:
      raise ValueError('{!r} is not a valid namespace. Available: {}.'.format(
          namespace, schema.FINDABLE_NAMESPACES))
    if constants.PREFIX_SEPARATOR in identifier:
      scope_name = identifier.split(constants.PREFIX_SEPARATOR)[0]
      try:
        attachment = self.namescope.get('attached_model', scope_name)
        found_element = attachment.find(
            namespace, identifier[(len(scope_name) + 1):])
      except (KeyError, ValueError):
        found_element = None
    else:
      try:
        found_element = self.namescope.get(namespace, identifier)
      except KeyError:
        found_element = None
      if found_element and self._parent:
        next_parent = found_element.parent
        while next_parent and next_parent != self:
          next_parent = next_parent.parent
        if not next_parent:
          found_element = None
    return found_element

  def find_all(self, namespace,
               immediate_children_only=False, exclude_attachments=False):
    """Finds all elements of a particular kind.

    The `namespace` argument specifies the kind of element to
    find. In most cases, this corresponds to the element's XML tag name.
    However, if an element has multiple specialized tags, then the namespace
    corresponds to the tag name of the most general element of that kind.
    For example, `namespace='joint'` would search for `<joint>` and
    `<freejoint>`, while `namespace='actuator'` would search for `<general>`,
    `<motor>`, `<position>`, `<velocity>`, and `<cylinder>`.

    Args:
      namespace: A string specifying the namespace being searched. See the
        docstring above for explanation.
      immediate_children_only: (optional) A boolean, if `True` then only
        the immediate children of this element are returned.
      exclude_attachments: (optional) A boolean, if `True` then elements
        belonging to attached models are excluded.

    Returns:
      A list of `mjcf.Element`.

    Raises:
      ValueError: if `namespace` is not a valid namespace.
    """
    if namespace not in schema.FINDABLE_NAMESPACES:
      raise ValueError('{!r} is not a valid namespace. Available: {}'.format(
          namespace, schema.FINDABLE_NAMESPACES))
    out = []
    children = self._children if exclude_attachments else self.all_children()
    for child in children:
      if (namespace == child.spec.namespace or
          # Direct children of attachment frames have custom namespaces of the
          # form "joint@attachment_frame_<id>".
          child.spec.namespace and child.spec.namespace.startswith(
              namespace + constants.NAMESPACE_SEPARATOR) or
          # Attachment frames are considered part of the "body" namespace.
          namespace == constants.BODY and isinstance(child, _AttachmentFrame)):
        out.append(child)
      if not immediate_children_only:
        out.extend(child.find_all(namespace,
                                  exclude_attachments=exclude_attachments))
    return out

  def enter_scope(self, scope_identifier):
    """Finds the root element of the given scope and returns it.

    This function allows the access to a nested scope that is a child of this
    element. The `scope_identifier` argument specifies the path to the child
    scope element.

    Args:
      scope_identifier: The path of the desired scope element.

    Returns:
      An `mjcf.Element` object, or `None` if a scope element with the
      specified path is not found.
    """
    if constants.PREFIX_SEPARATOR in scope_identifier:
      scope_name = scope_identifier.split(constants.PREFIX_SEPARATOR)[0]
      try:
        attachment = self.namescope.get('attached_model', scope_name)
      except KeyError:
        return None

      scope_suffix = scope_identifier[(len(scope_name) + 1):]
      if scope_suffix:
        return attachment.enter_scope(scope_suffix)
      else:
        return attachment
    else:
      try:
        return self.namescope.get('attached_model', scope_identifier)
      except KeyError:
        return None

  def _check_valid_attribute(self, attribute_name):
    if attribute_name not in self._spec.attributes:
      raise AttributeError(
          '{!r} is not a valid attribute for <{}>'.format(
              attribute_name, self._spec.name))

  def _get_attribute(self, attribute_name):
    self._check_valid_attribute(attribute_name)
    return self._attributes[attribute_name].value

  def get_attribute_xml_string(self, attribute_name, prefix_root=None):
    self._check_valid_attribute(attribute_name)
    return self._attributes[attribute_name].to_xml_string(prefix_root)

  def get_attributes(self):
    fix_attribute_name = (
        lambda name: constants.DCLASS if name == constants.CLASS else name)
    return collections.OrderedDict(
        [(fix_attribute_name(name), self._get_attribute(name))
         for name in six.iterkeys(self._spec.attributes)
         if self._get_attribute(name) is not None])

  def _set_attribute(self, attribute_name, value):
    self._check_valid_attribute(attribute_name)
    self._attributes[attribute_name].value = value
    self.namescope.increment_revision()

  def set_attributes(self, **kwargs):
    if constants.DCLASS in kwargs:
      kwargs[constants.CLASS] = kwargs[constants.DCLASS]
      del kwargs[constants.DCLASS]
    old_values = []
    with debugging.freeze_current_stack_trace():
      for attribute_name, new_value in six.iteritems(kwargs):
        old_value = self._get_attribute(attribute_name)
        try:
          self._set_attribute(attribute_name, new_value)
          old_values.append((attribute_name, old_value))
        except:  # pylint: disable=bare-except
          # On failure, restore old attribute values for those already set.
          for name, old_value in old_values:
            self._set_attribute(name, old_value)
          # Then raise a meaningful error.
          err_type, err, tb = sys.exc_info()
          message = (
              'during assignment to attribute {!r} of element <{}>: {}'
              .format(attribute_name, self._spec.name, err))
          six.reraise(err_type, err_type(message), tb)

  def _remove_attribute(self, attribute_name):
    self._check_valid_attribute(attribute_name)
    self._attributes[attribute_name].clear()
    self.namescope.increment_revision()

  def _check_valid_child(self, element_name):
    try:
      return self._spec.children[element_name]
    except KeyError:
      raise AttributeError(
          '<{}> is not a valid child of <{}>'
          .format(element_name, self._spec.name))

  def get_children(self, element_name):
    child_spec = self._check_valid_child(element_name)
    if child_spec.repeated:
      return _ElementListView(spec=child_spec, parent=self)
    else:
      for child in self._children:
        if child.tag == element_name:
          return child
      if child_spec.on_demand:
        return None
      else:
        raise RuntimeError(
            'Cannot find the non-repeated child <{}> of <{}>. '
            'This should never happen, as we pre-create these in __init__. '
            'Please file an bug report. Thank you.'
            .format(element_name, self._spec.name))

  def add(self, element_name, **kwargs):
    """Add a new child element to this element.

    Args:
      element_name: The tag of the element to add.
      **kwargs: Attributes of the new element being created.

    Raises:
      ValueError: If the 'element_name' is not a valid child, or if an invalid
        attribute is specified in `kwargs`.

    Returns:
      An `mjcf.Element` corresponding to the newly created child element.
    """
    child_spec = self._check_valid_child(element_name)
    if child_spec.on_demand:
      need_new_on_demand = self.get_children(element_name) is None
    else:
      need_new_on_demand = False
    if not (child_spec.repeated or need_new_on_demand):
      raise ValueError('A <{}> child already exists, please access it directly.'
                       .format(element_name))
    new_element = _make_element(child_spec, self, attributes=kwargs)
    self._children.append(new_element)
    self.namescope.increment_revision()
    return new_element

  def __getattr__(self, name):
    if self._last_attribute_error:
      # This means that we got here through a @property raising AttributeError.
      # We therefore just re-raise the last AttributeError back to the user.
      # Note that self._last_attribute_error was set by our specially
      # instrumented @property decorator.
      exc_info = self._last_attribute_error
      self._last_attribute_error = None
      six.reraise(*exc_info)  # pylint: disable=not-an-iterable
    elif name in self._spec.children:
      return self.get_children(name)
    elif name in self._spec.attributes:
      return self._get_attribute(name)
    elif name == constants.DCLASS and constants.CLASS in self._spec.attributes:
      return self._get_attribute(constants.CLASS)
    else:
      raise AttributeError('object has no attribute: {}'.format(name))

  def __setattr__(self, name, value):
    # If this name corresponds to a descriptor for a slotted attribute or
    # settable property then try to invoke the descriptor to set the attribute
    # and return if successful.
    klass_attr = getattr(type(self), name, None)
    if klass_attr is not None:
      try:
        return klass_attr.__set__(self, value)
      except AttributeError:
        pass
    # If we did not find a settable descriptor then we look in the attribute
    # spec to see if there is a MuJoCo attribute matching this name.
    attribute_name = name if name != constants.DCLASS else constants.CLASS
    if attribute_name in self._spec.attributes:
      self._set_attribute(attribute_name, value)
    else:
      raise AttributeError('can\'t set attribute: {}'.format(name))

  def __delattr__(self, name):
    if name in self._spec.children:
      if self._spec.children[name].repeated:
        raise AttributeError(
            '`{0}` is a collection of child elements, '
            'which cannot be deleted. Did you mean to call `{0}.clear()`?'
            .format(name))
      else:
        return self.get_children(name).remove()
    elif name in self._spec.attributes:
      return self._remove_attribute(name)
    else:
      raise AttributeError('object has no attribute: {}'.format(name))

  def _check_attachments_on_remove(self, affect_attachments):
    if not affect_attachments and self._attachments:
      raise ValueError(
          'please use remove(affect_attachments=True) as this will affect some '
          'attributes and/or children belonging to an attached model')
    for child in self._children:
      child._check_attachments_on_remove(affect_attachments)  # pylint: disable=protected-access

  def remove(self, affect_attachments=False):
    """Removes this element from the model."""
    self._check_attachments_on_remove(affect_attachments)
    if affect_attachments:
      for attachment in six.itervalues(self._attachments):
        attachment.remove(affect_attachments=True)
    for child in list(self._children):
      child.remove(affect_attachments)
    if self._spec.repeated or self._spec.on_demand:
      self._parent._children.remove(self)  # pylint: disable=protected-access
      for attribute in six.itervalues(self._attributes):
        attribute._force_clear()  # pylint: disable=protected-access
      self._parent = None
      self._is_removed = True
    else:
      for attribute in six.itervalues(self._attributes):
        attribute._force_clear()  # pylint: disable=protected-access
    self.namescope.increment_revision()

  @property
  def is_removed(self):
    return self._is_removed

  def all_children(self):
    all_children = [child for child in self._children]
    for attachment in six.itervalues(self._attachments):
      all_children += [child for child in attachment.all_children()
                       if child.spec.repeated]
    return all_children

  def to_xml(self, prefix_root=None, debug_context=None):
    """Generates an etree._Element corresponding to this MJCF element.

    Args:
      prefix_root: (optional) A `NameScope` object to be treated as root
        for the purpose of calculating the prefix.
        If `None` then no prefix is included.
      debug_context: (optional) A `debugging.DebugContext` object to which
        the debugging information associated with the generated XML is written.
        This is intended for internal use within PyMJCF; users should never need
        manually pass this argument.

    Returns:
      An etree._Element object.
    """
    prefix_root = prefix_root or self.namescope
    xml_element = etree.Element(self._spec.name)
    self._attributes_to_xml(xml_element, prefix_root, debug_context)
    self._children_to_xml(xml_element, prefix_root, debug_context)
    return xml_element

  def _attributes_to_xml(self, xml_element, prefix_root, debug_context=None):
    del debug_context  # Unused.
    for attribute_name, attribute in six.iteritems(self._attributes):
      attribute_value = attribute.to_xml_string(prefix_root)
      if attribute_name == self._spec.identifier and attribute_value is None:
        xml_element.set(attribute_name, self.full_identifier)
      elif attribute_value is None:
        continue
      else:
        xml_element.set(attribute_name, attribute_value)

  def _children_to_xml(self, xml_element, prefix_root, debug_context=None):
    for child in self.all_children():
      child_xml = child.to_xml(prefix_root, debug_context)
      if (child_xml.attrib or len(child_xml)  # pylint: disable=g-explicit-length-test
          or child.spec.repeated or child.spec.on_demand):
        xml_element.append(child_xml)
        if debugging.debug_mode() and debug_context:
          debug_comment = debug_context.register_element_for_debugging(child)
          xml_element.append(debug_comment)
          if len(child_xml) > 0:  # pylint: disable=g-explicit-length-test
            child_xml.insert(0, copy.deepcopy(debug_comment))

  def to_xml_string(self, prefix_root=None,
                    self_only=False, pretty_print=True, debug_context=None):
    """Generates an XML string corresponding to this MJCF element.

    Args:
      prefix_root: (optional) A `NameScope` object to be treated as root
        for the purpose of calculating the prefix.
        If `None` then no prefix is included.
      self_only: (optional) A boolean, whether to generate an XML corresponding
        only to this element without any children.
      pretty_print: (optional) A boolean, whether to the XML string should be
        properly indented.
      debug_context: (optional) A `debugging.DebugContext` object to which
        the debugging information associated with the generated XML is written.
        This is intended for internal use within PyMJCF; users should never need
        manually pass this argument.

    Returns:
      A string.
    """
    xml_element = self.to_xml(prefix_root, debug_context)
    if self_only and len(xml_element) > 0:  # pylint: disable=g-explicit-length-test
      etree.strip_elements(xml_element, '*')
      xml_element.text = '...'
    if (self_only and self._spec.identifier and
        not self._attributes[self._spec.identifier].to_xml_string(prefix_root)):
      del xml_element.attrib[self._spec.identifier]
    xml_string = util.to_native_string(
        etree.tostring(xml_element, pretty_print=pretty_print))
    if pretty_print and debug_context:
      return debug_context.commit_xml_string(xml_string)
    else:
      return xml_string

  def __str__(self):
    return self.to_xml_string(self_only=True, pretty_print=False)

  def __repr__(self):
    return 'MJCF Element: ' + str(self)

  def _check_valid_attachment(self, other):
    self_spec = self._spec
    if self_spec.name == constants.WORLDBODY:
      self_spec = self._spec.children[constants.BODY]

    other_spec = other.spec
    if other_spec.name == constants.WORLDBODY:
      other_spec = other_spec.children[constants.BODY]

    if other_spec != self_spec:
      raise ValueError(
          'The attachment must have the same spec as this element.')

  def _attach(self, other, exclude_worldbody=False, dry_run=False):
    """Attaches another element of the same spec to this element.

    All children of `other` will be treated as children of this element.
    All XML attributes which are defined in `other` but not defined in this
    element will be copied over, and any conflicting XML attribute value causes
    an error. After the attachment, any XML attribute modified in this element
    will also affect `other` and vice versa.

    Children of this element which are not a repeated elements will also be
    attached by the corresponding children of `other`.

    Args:
      other: Another Element with the same spec.
      exclude_worldbody: (optional) A boolean. If `True`, then don't do anything
        if `other` is a worldbody.
      dry_run: (optional) A boolean, if `True` only verify that the operation
        is valid without actually committing any change.

    Raises:
      ValueError: If `other` has a different spec, or if there are conflicting
        XML attribute values.
    """
    self._check_valid_attachment(other)
    if exclude_worldbody and other.tag == constants.WORLDBODY:
      return
    if dry_run:
      self._check_conflicting_attributes(other, copying=False)
    else:
      self._attachments[other.namescope] = other
      self._sync_attributes(other, copying=False)
    self._attach_children(other, exclude_worldbody, dry_run)
    if other.tag != constants.WORLDBODY and not dry_run:
      other._alias_attributes_dict(self._attributes)  # pylint: disable=protected-access

  def _detach(self, other_namescope):
    """Detaches a model with the specified namescope."""
    attached_element = self._attachments.get(other_namescope)
    if attached_element:
      attached_element._restore_attributes_dict()  # pylint: disable=protected-access
      del self._attachments[other_namescope]
    for child in self._children:
      child._detach(other_namescope)  # pylint: disable=protected-access

  def _check_conflicting_attributes(self, other, copying):
    for attribute_name, other_attribute in six.iteritems(
        other.get_attributes()):
      if attribute_name == constants.DCLASS:
        attribute_name = constants.CLASS
      if ((not self._attributes[attribute_name].conflict_allowed)
          and self._attributes[attribute_name].value is not None
          and other_attribute is not None
          and np.asarray(
              self._attributes[attribute_name].value != other_attribute).any()):
        raise ValueError(
            'Conflicting values for attribute `{}`: {} vs {}'
            .format(attribute_name,
                    self._attributes[attribute_name].value,
                    other_attribute))

  def _sync_attributes(self, other, copying):
    self._check_conflicting_attributes(other, copying)
    for attribute_name, other_attribute in six.iteritems(
        other.get_attributes()):
      if attribute_name == constants.DCLASS:
        attribute_name = constants.CLASS

      self_attribute = self._attributes[attribute_name]
      if other_attribute is not None:
        if self_attribute.conflict_behavior in _CONFLICT_BEHAVIOR_FUNC:
          if self_attribute.value is not None:
            self_attribute.value = (
                _CONFLICT_BEHAVIOR_FUNC[self_attribute.conflict_behavior](
                    self_attribute.value, other_attribute))
          else:
            self_attribute.value = other_attribute
        elif copying or not self_attribute.conflict_allowed:
          self_attribute.value = other_attribute

  def _attach_children(self, other, exclude_worldbody, dry_run=False):
    for other_child in other.all_children():
      if not other_child.spec.repeated:
        self_child = self.get_children(other_child.spec.name)
        self_child._attach(other_child, exclude_worldbody, dry_run)  # pylint: disable=protected-access

  def resolve_references(self):
    for attribute in six.itervalues(self._attributes):
      if isinstance(attribute, attribute_types.Reference):
        if attribute.value and isinstance(attribute.value, six.string_types):
          referred = self.root.find(
              attribute.reference_namespace, attribute.value)
          if referred:
            attribute.value = referred
    for child in self.all_children():
      child.resolve_references()

  def _update_references(self, reference_dict):
    for attribute in six.itervalues(self._attributes):
      if isinstance(attribute, attribute_types.Reference):
        if attribute.value in reference_dict:
          attribute.value = reference_dict[attribute.value]
    for child in self.all_children():
      child._update_references(reference_dict)  # pylint: disable=protected-access


class _AttachableElement(_ElementImpl):
  """Specialized object representing a <site> or <worldbody> element.

  This element defines a frame to which another MJCF model can be attached.
  """
  __slots__ = []

  def attach(self, attachment):
    """Attaches another MJCF model at this site.

    An empty <body> will be created as an attachment frame. All children of
    `attachment`'s <worldbody> will be treated as children of this frame.
    Furthermore, all other elements in `attachment` are merged into the root
    of the MJCF model to which this element belongs.

    Args:
      attachment: An MJCF `RootElement`

    Returns:
      An `mjcf.Element` corresponding to the attachment frame. A joint can be
      added directly to this frame to give degrees of freedom to the attachment.

    Raises:
      ValueError: If `other` is not a valid attachment to this element.
    """
    if not isinstance(attachment, RootElement):
      raise ValueError('Expected a mjcf.RootElement: got {}'
                       .format(attachment))
    if attachment.namescope.parent is not None:
      raise ValueError('The model specified is already attached elsewhere')
    if attachment.namescope == self.namescope:
      raise ValueError('Cannot merge a model to itself')
    self.root._attach(attachment, exclude_worldbody=True, dry_run=True)  # pylint: disable=protected-access

    if self.namescope.has_identifier('namescope', attachment.model):
      id_number = 1
      while self.namescope.has_identifier(
          'namescope', '{}_{}'.format(attachment.model, id_number)):
        id_number += 1
      attachment.model = '{}_{}'.format(attachment.model, id_number)
    attachment.namescope.parent = self.namescope

    if self.tag == constants.WORLDBODY:
      frame_parent = self
      frame_siblings = self._children
      index = len(frame_siblings)
    else:
      frame_parent = self._parent
      frame_siblings = self._parent._children  # pylint: disable=protected-access
      index = frame_siblings.index(self) + 1
      while (index < len(frame_siblings)
             and isinstance(frame_siblings[index], _AttachmentFrame)):
        index += 1
    frame = _AttachmentFrame(frame_parent, self, attachment)
    frame_siblings.insert(index, frame)
    self.root._attach(attachment, exclude_worldbody=True)  # pylint: disable=protected-access
    return frame


class _AttachmentFrame(_ElementImpl):
  """An specialized <body> representing a frame holding an external attachment.
  """
  __slots__ = ['_site', '_attachment']

  def __init__(self, parent, site, attachment):
    if parent.tag == constants.WORLDBODY:
      spec = schema.WORLD_ATTACHMENT_FRAME
    else:
      spec = schema.ATTACHMENT_FRAME

    spec_is_copied = False
    for child_name, child_spec in six.iteritems(spec.children):
      if child_spec.namespace:
        if not spec_is_copied:
          spec = copy.deepcopy(spec)
          spec_is_copied = True
        spec_as_dict = child_spec._asdict()
        spec_as_dict['namespace'] = '{}{}attachment_frame_{}'.format(
            child_spec.namespace, constants.NAMESPACE_SEPARATOR, id(self))
        spec.children[child_name] = type(child_spec)(**spec_as_dict)

    attributes = {}
    with debugging.freeze_current_stack_trace():
      for attribute_name in six.iterkeys(spec.attributes):
        if hasattr(site, attribute_name):
          attributes[attribute_name] = getattr(site, attribute_name)
    super(_AttachmentFrame, self).__init__(spec, parent, attributes)
    self._site = site
    self._attachment = attachment
    self._attachments[attachment.namescope] = attachment.worldbody
    self.namescope.add('attachment_frame', attachment.namescope.name, self)
    self.namescope.add('attached_model', attachment.namescope.name, attachment)

  def prefixed_identifier(self, prefix_root=None):
    prefix = self.namescope.full_prefix(prefix_root)
    return prefix + self._attachment.namescope.name + constants.PREFIX_SEPARATOR

  def to_xml(self, prefix_root=None, debug_context=None):
    xml_element = (
        super(_AttachmentFrame, self).to_xml(prefix_root, debug_context))
    xml_element.set('name', self.prefixed_identifier(prefix_root))
    return xml_element

  @property
  def full_identifier(self):
    return self.prefixed_identifier(self.namescope.root)

  def _detach(self, other_namescope):
    super(_AttachmentFrame, self)._detach(other_namescope)
    if other_namescope is self._attachment.namescope:
      self.namescope.remove('attachment_frame', self._attachment.namescope.name)
      self.namescope.remove('attached_model', self._attachment.namescope.name)
      self.remove()


class _AttachmentFrameChild(_ElementImpl):
  """A child element of an attachment frame.

  Right now, this is always a <joint> or a <freejoint>. The name of the joint
  is not freely specifiable, but instead just inherits from the parent frame.
  This ensures uniqueness, as attachment frame identifiers always end in '/'.
  """
  __slots__ = []

  def to_xml(self, prefix_root=None, debug_context=None):
    xml_element = (
        super(_AttachmentFrameChild, self).to_xml(prefix_root, debug_context))
    if self.spec.namespace is not None:
      if self.name:
        name = (self._parent.prefixed_identifier(prefix_root) +
                self.name + constants.PREFIX_SEPARATOR)
      else:
        name = self._parent.prefixed_identifier(prefix_root)
      xml_element.set('name', name)
    return xml_element

  def prefixed_identifier(self, prefix_root=None):
    if self.name:
      return (self._parent.prefixed_identifier(prefix_root) +
              self.name + constants.PREFIX_SEPARATOR)
    else:
      return self._parent.prefixed_identifier(prefix_root)


class _DefaultElement(_ElementImpl):
  """Specialized object representing a <default> element.

  This is necessary for the proper handling of global defaults.
  """
  __slots__ = []

  def _attach(self, other, exclude_worldbody=False, dry_run=False):
    self._check_valid_attachment(other)
    if ((not isinstance(self._parent, RootElement))
        or (not isinstance(other.parent, RootElement))):
      raise ValueError('Only global <{}> can be attached'
                       .format(constants.DEFAULT))
    if not dry_run:
      self._attachments[other.namescope] = other

  def all_children(self):
    return [child for child in self._children]

  def to_xml(self, prefix_root=None, debug_context=None):
    prefix_root = prefix_root or self.namescope
    xml_element = (
        super(_DefaultElement, self).to_xml(prefix_root, debug_context))
    if isinstance(self._parent, RootElement):
      root_default = etree.Element(self._spec.name)
      root_default.append(xml_element)
      for attachment in six.itervalues(self._attachments):
        attachment_xml = attachment.to_xml(prefix_root, debug_context)
        for attachment_child_xml in attachment_xml:
          root_default.append(attachment_child_xml)
      xml_element = root_default
    return xml_element


class _ActuatorElement(_ElementImpl):
  """Specialized object representing an <actuator> element.

  This is necessary because MuJoCo requires that all 3rd-order actuators (i.e.
  those with internal dynamics) come after all 2nd-order actuators in the
  generated XML.
  """
  __slots__ = ()

  def _is_third_order_actuator(self, child):
    if child.tag == 'general':
      return child.dyntype and child.dyntype != 'none'
    elif child.tag == 'cylinder':
      return True  # The `<cylinder>` shortcut has 'filter' dynamics.
    else:
      return False  # No other actuator shortcuts have internal dynamics.

  def _children_to_xml(self, xml_element, prefix_root, debug_context=None):
    second_order = []
    third_order = []
    debug_comments = {}
    for child in self.all_children():
      child_xml = child.to_xml(prefix_root, debug_context)
      if (child_xml.attrib or len(child_xml)  # pylint: disable=g-explicit-length-test
          or child.spec.repeated or child.spec.on_demand):
        if self._is_third_order_actuator(child):
          third_order.append(child_xml)
        else:
          second_order.append(child_xml)
        if debugging.debug_mode() and debug_context:
          debug_comment = debug_context.register_element_for_debugging(child)
          debug_comments[child_xml] = debug_comment
          if len(child_xml) > 0:  # pylint: disable=g-explicit-length-test
            child_xml.insert(0, copy.deepcopy(debug_comment))
    # Ensure that all second-order actuators come before third-order actuators
    # in the XML.
    for child_xml in second_order + third_order:
      xml_element.append(child_xml)
      if debugging.debug_mode() and debug_context:
        xml_element.append(debug_comments[child_xml])


class RootElement(_ElementImpl):
  """The root `<mujoco>` element of an MJCF model."""
  __slots__ = ['_namescope']

  def __init__(self, model=None, model_dir='', assets=None):
    model = model or 'unnamed_model'
    self._namescope = namescope.NameScope(
        model, self, model_dir=model_dir, assets=assets)
    super(RootElement, self).__init__(
        spec=schema.MUJOCO, parent=None, attributes={'model': model})

  def _attach(self, other, exclude_worldbody=False, dry_run=False):
    self._check_valid_attachment(other)
    if not dry_run:
      self._attachments[other.namescope] = other
    self._attach_children(other, exclude_worldbody, dry_run)
    self.namescope.increment_revision()

  @property
  def namescope(self):
    return self._namescope

  @property
  def root(self):
    return self

  @property
  def model(self):
    return self._namescope.name

  @model.setter
  def model(self, new_name):
    old_name = self._namescope.name
    self._namescope.name = new_name
    self._attributes['model'].value = new_name
    if self.parent_model:
      self.parent_model.namescope.rename('attachment_frame', old_name, new_name)
      self.parent_model.namescope.rename('attached_model', old_name, new_name)

  def attach(self, other):
    return self.worldbody.attach(other)

  def detach(self):
    parent_model = self.parent_model
    if not parent_model:
      raise RuntimeError(
          'Cannot `detach` a model that is not attached to some other model.')
    else:
      parent_model._detach(self.namescope)  # pylint: disable=protected-access
      self.namescope.parent = None

  def include_copy(self, other, override_attributes=False):
    other_copier = copier.Copier(other)
    new_elements = other_copier.copy_into(self, override_attributes)
    self._update_references(new_elements)
    self.namescope.increment_revision()

  @property
  def parent_model(self):
    """The RootElement of the MJCF model to which this one is attached."""
    namescope_parent = self._namescope.parent
    return namescope_parent.mjcf_model if namescope_parent else None

  @property
  def root_model(self):
    return self.parent_model.root_model if self.parent_model else self

  def get_assets(self):
    """Returns a dict containing the binary assets referenced in this model.

    This will contain `{vfs_filename: contents}` pairs. `vfs_filename` will be
    the name of the asset in MuJoCo's Virtual File System, which corresponds to
    the filename given in the XML returned by `to_xml_string()`. `contents` is a
    bytestring.

    This dict can be used together with the result of `to_xml_string()` to
    construct a `mujoco.Physics` instance:

    ```python
    physics = mujoco.Physics.from_xml_string(
        xml_string=mjcf_model.to_xml_string(),
        assets=mjcf_model.get_assets())
    ```
    """
    # Get the assets referenced within this `RootElement`'s namescope.
    assets = {file_obj.to_xml_string(): file_obj.get_contents()
              for file_obj in self.namescope.files
              if file_obj.value}

    # Recursively add assets belonging to attachments.
    for attached_model in six.itervalues(self._attachments):
      assets.update(attached_model.get_assets())

    return assets

  @property
  def full_identifier(self):
    return self._namescope.full_prefix(self._namescope.root)

  def __copy__(self):
    new_model = RootElement(model=self._namescope.name,
                            model_dir=self.namescope.model_dir)
    new_model.include_copy(self)
    return new_model

  def __deepcopy__(self, _):
    return self.__copy__()

  def is_same_as(self, other):
    if other is None or other.spec != self._spec:
      return False
    return self._has_same_children_as(other)


class _ElementListView(object):
  """A hybrid list/dict-like view to a group of repeated MJCF elements."""

  def __init__(self, spec, parent):
    self._spec = spec
    self._parent = parent
    # pylint: disable=protected-access
    self._elements = self._parent._children
    self._scoped_elements = collections.OrderedDict(
        [(scope_namescope.name, getattr(scoped_parent, self._spec.name))
         for scope_namescope, scoped_parent
         in six.iteritems(self._parent._attachments)])

  @property
  def spec(self):
    return self._spec

  @property
  def tag(self):
    return self._spec.name

  @property
  def namescope(self):
    return self._parent.namescope

  @property
  def parent(self):
    return self._parent

  def __len__(self):
    return len(self._full_list())

  def __iter__(self):
    return iter(self._full_list())

  def _identifier_not_found_error(self, index):
    return KeyError('An element <{}> with {}={!r} does not exist'
                    .format(self._spec.name, self._spec.identifier, index))

  def _find_index(self, index):
    """Locates an element given the index among siblings with the same tag."""
    if isinstance(index, str) and self._spec.identifier:
      for i, element in enumerate(self._elements):
        if (element.tag == self._spec.name
            and getattr(element, self._spec.identifier) == index):
          return i
      raise self._identifier_not_found_error(index)
    else:
      count = 0
      for i, element in enumerate(self._elements):
        if element.tag == self._spec.name:
          if index == count:
            return i
          else:
            count += 1
      raise IndexError('list index out of range')

  def _full_list(self):
    out_list = [element for element in self._elements
                if element.tag == self._spec.name]
    for scoped_elements in six.itervalues(self._scoped_elements):
      out_list += scoped_elements[:]
    return out_list

  def clear(self):
    for child in self._full_list():
      child.remove()

  def __getitem__(self, index):
    if (isinstance(index, str) and self._spec.identifier
        and constants.PREFIX_SEPARATOR in index):
      scope_name = index.split(constants.PREFIX_SEPARATOR)[0]
      scoped_elements = self._scoped_elements[scope_name]
      try:
        return scoped_elements[index[(len(scope_name) + 1):]]
      except KeyError:
        # Re-raise so that the error shows the full, un-stripped index string
        raise self._identifier_not_found_error(index)
    elif isinstance(index, slice) or (isinstance(index, int) and index < 0):
      return self._full_list()[index]
    else:
      return self._elements[self._find_index(index)]

  def __delitem__(self, index):
    found_index = self._find_index(index)
    self._elements[found_index].remove()

  def __str__(self):
    return str(
        [element.to_xml_string(
            prefix_root=self.namescope, self_only=True, pretty_print=False)
         for element in self._full_list()])

  def __repr__(self):
    return 'MJCF Elements List: ' + str(self)


# This restores @property back to Python's built-in one.
del property
del _raw_property



# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Global constants used in manipulation tasks."""

CONTROL_TIMESTEP = 0.04  # Interval between agent actions, in seconds.

# Predefined RGBA values
RED = (1., 0., 0., 0.3)
GREEN = (0., 1., 0., 0.3)
BLUE = (0., 0., 1., 0.3)
CYAN = (0., 1., 1., 0.3)
MAGENTA = (1., 0., 1., 0.3)
YELLOW = (1., 1., 0., 0.3)

TASK_SITE_GROUP = 3  # Invisible group for task-related sites.



# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Control suite environments explorer."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl import app
from absl import flags
from dm_control import suite
from dm_control.suite.wrappers import action_noise
from six.moves import input

from dm_control import viewer


_ALL_NAMES = ['.'.join(domain_task) for domain_task in suite.ALL_TASKS]

flags.DEFINE_enum('environment_name', None, _ALL_NAMES,
                  'Optional \'domain_name.task_name\' pair specifying the '
                  'environment to load. If unspecified a prompt will appear to '
                  'select one.')
flags.DEFINE_bool('timeout', True, 'Whether episodes should have a time limit.')
flags.DEFINE_bool('visualize_reward', True,
                  'Whether to vary the colors of geoms according to the '
                  'current reward value.')
flags.DEFINE_float('action_noise', 0.,
                   'Standard deviation of Gaussian noise to apply to actions, '
                   'expressed as a fraction of the max-min range for each '
                   'action dimension. Defaults to 0, i.e. no noise.')
FLAGS = flags.FLAGS


def prompt_environment_name(prompt, values):
  environment_name = None
  while not environment_name:
    environment_name = input(prompt)
    if not environment_name or values.index(environment_name) < 0:
      print('"%s" is not a valid environment name.' % environment_name)
      environment_name = None
  return environment_name


def main(argv):
  del argv
  environment_name = FLAGS.environment_name
  if environment_name is None:
    print('\n  '.join(['Available environments:'] + _ALL_NAMES))
    environment_name = prompt_environment_name(
        'Please select an environment name: ', _ALL_NAMES)

  index = _ALL_NAMES.index(environment_name)
  domain_name, task_name = suite.ALL_TASKS[index]

  task_kwargs = {}
  if not FLAGS.timeout:
    task_kwargs['time_limit'] = float('inf')

  def loader():
    env = suite.load(
        domain_name=domain_name, task_name=task_name, task_kwargs=task_kwargs)
    env.task.visualize_reward = FLAGS.visualize_reward
    if FLAGS.action_noise > 0:
      env = action_noise.Wrapper(env, scale=FLAGS.action_noise)
    return env

  viewer.launch(loader)


if __name__ == '__main__':
  app.run(main)



# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests for base variation operations."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import operator
from absl.testing import absltest
from absl.testing import parameterized
from dm_control.composer import variation
from dm_control.composer.variation import deterministic
import numpy as np


class VariationTest(parameterized.TestCase):

  def setUp(self):
    super(VariationTest, self).setUp()
    self.value_1 = 3
    self.variation_1 = deterministic.Constant(self.value_1)
    self.value_2 = 5
    self.variation_2 = deterministic.Constant(self.value_2)

  @parameterized.parameters(['neg'])
  def test_unary_operator(self, name):
    func = getattr(operator, name)
    self.assertEqual(
        variation.evaluate(func(self.variation_1)),
        func(self.value_1))

  @parameterized.parameters(['add', 'sub', 'mul', 'truediv', 'floordiv', 'pow'])
  def test_binary_operator(self, name):
    func = getattr(operator, name)
    self.assertEqual(
        variation.evaluate(func(self.value_1, self.variation_2)),
        func(self.value_1, self.value_2))
    self.assertEqual(
        variation.evaluate(func(self.variation_1, self.value_2)),
        func(self.value_1, self.value_2))
    self.assertEqual(
        variation.evaluate(func(self.variation_1, self.variation_2)),
        func(self.value_1, self.value_2))

  def test_getitem(self):
    value = deterministic.Constant(np.array([4, 5, 6, 7, 8]))
    np.testing.assert_array_equal(
        variation.evaluate(value[[3, 1]]),
        [7, 5])

if __name__ == '__main__':
  absltest.main()



# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Locomotion texture assets."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import os
import sys

ROOT_DIR = '../locomotion/arenas/assets'


def get_texturedir(style):
  return os.path.join(ROOT_DIR, style)

SKY_STYLES = ('outdoor_natural')

SkyBox = collections.namedtuple(
    'SkyBox', ('file', 'gridsize', 'gridlayout'))


def get_sky_texture_info(style):
  if style not in SKY_STYLES:
    raise ValueError('`style` should be one of {}: got {!r}'.format(
        SKY_STYLES, style))
  return SkyBox(file='OutdoorSkybox2048.png',
                gridsize='3 4',
                gridlayout='.U..LFRB.D..')


GROUND_STYLES = ('outdoor_natural')

GroundTexture = collections.namedtuple(
    'GroundTexture', ('file', 'type'))


def get_ground_texture_info(style):
  if style not in GROUND_STYLES:
    raise ValueError('`style` should be one of {}: got {!r}'.format(
        GROUND_STYLES, style))
  return GroundTexture(
      file='OutdoorGrassFloorD.png',
      type='2d')







# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Position & velocity actuators whose controls are scaled to a given range."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

_DISALLOWED_KWARGS = frozenset(
    ['biastype', 'gainprm', 'biasprm', 'ctrllimited',
     'joint', 'tendon', 'site', 'slidersite', 'cranksite'])
_ALLOWED_TAGS = frozenset(['joint', 'tendon', 'site'])

_GOT_INVALID_KWARGS = 'Received invalid keyword argument(s): {}'
_GOT_INVALID_TARGET = '`target` tag type should be one of {}: got {{}}'.format(
    sorted(_ALLOWED_TAGS))


def _check_target_and_kwargs(target, **kwargs):
  invalid_kwargs = _DISALLOWED_KWARGS.intersection(kwargs)
  if invalid_kwargs:
    raise TypeError(_GOT_INVALID_KWARGS.format(sorted(invalid_kwargs)))
  if target.tag not in _ALLOWED_TAGS:
    raise TypeError(_GOT_INVALID_TARGET.format(target))


def add_position_actuator(target, qposrange, ctrlrange=(-1, 1),
                          kp=1.0, **kwargs):
  """Adds a scaled position actuator that is bound to the specified element.

  This is equivalent to MuJoCo's built-in `<position>` actuator where an affine
  transformation is pre-applied to the control signal, such that the minimum
  control value corresponds to the minimum desired position, and the
  maximum control value corresponds to the maximum desired position.

  Args:
    target: A PyMJCF joint, tendon, or site element object that is to be
      controlled.
    qposrange: A sequence of two numbers specifying the allowed range of target
      position.
    ctrlrange: A sequence of two numbers specifying the allowed range of
      this actuator's control signal.
    kp: The gain parameter of this position actuator.
    **kwargs: Additional MJCF attributes for this actuator element.
      The following attributes are disallowed: `['biastype', 'gainprm',
      'biasprm', 'ctrllimited', 'joint', 'tendon', 'site',
      'slidersite', 'cranksite']`.

  Returns:
    A PyMJCF actuator element that has been added to the MJCF model containing
    the specified `target`.

  Raises:
    TypeError: `kwargs` contains an unrecognized or disallowed MJCF attribute,
      or `target` is not an allowed MJCF element type.
  """
  _check_target_and_kwargs(target, **kwargs)
  kwargs[target.tag] = target

  slope = (qposrange[1] - qposrange[0]) / (ctrlrange[1] - ctrlrange[0])
  g0 = kp * slope
  b0 = kp * (qposrange[0] - slope * ctrlrange[0])
  b1 = -kp
  b2 = 0
  return target.root.actuator.add('general',
                                  biastype='affine',
                                  gainprm=[g0],
                                  biasprm=[b0, b1, b2],
                                  ctrllimited=True,
                                  ctrlrange=ctrlrange,
                                  **kwargs)


def add_velocity_actuator(target, qvelrange, ctrlrange=(-1, 1),
                          kv=1.0, **kwargs):
  """Adds a scaled velocity actuator that is bound to the specified element.

  This is equivalent to MuJoCo's built-in `<velocity>` actuator where an affine
  transformation is pre-applied to the control signal, such that the minimum
  control value corresponds to the minimum desired velocity, and the
  maximum control value corresponds to the maximum desired velocity.

  Args:
    target: A PyMJCF joint, tendon, or site element object that is to be
      controlled.
    qvelrange: A sequence of two numbers specifying the allowed range of target
      velocity.
    ctrlrange: A sequence of two numbers specifying the allowed range of
      this actuator's control signal.
    kv: The gain parameter of this velocity actuator.
    **kwargs: Additional MJCF attributes for this actuator element.
      The following attributes are disallowed: `['biastype', 'gainprm',
      'biasprm', 'ctrllimited', 'joint', 'tendon', 'site',
      'slidersite', 'cranksite']`.

  Returns:
    A PyMJCF actuator element that has been added to the MJCF model containing
    the specified `target`.

  Raises:
    TypeError: `kwargs` contains an unrecognized or disallowed MJCF attribute,
      or `target` is not an allowed MJCF element type.
  """
  _check_target_and_kwargs(target, **kwargs)
  kwargs[target.tag] = target

  slope = (qvelrange[1] - qvelrange[0]) / (ctrlrange[1] - ctrlrange[0])
  g0 = kv * slope
  b0 = kv * (qvelrange[0] - slope * ctrlrange[0])
  b1 = 0
  b2 = -kv
  return target.root.actuator.add('general',
                                  biastype='affine',
                                  gainprm=[g0],
                                  biasprm=[b0, b1, b2],
                                  ctrllimited=True,
                                  ctrlrange=ctrlrange,
                                  **kwargs)



# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""A task where the goal is to move the hand close to a target prop or site."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

from dm_control import composer
from dm_control.composer import initializers
from dm_control.composer.observation import observable
from dm_control.composer.variation import distributions
from dm_control.entities import props
from dm_control.manipulation.shared import arenas
from dm_control.manipulation.shared import cameras
from dm_control.manipulation.shared import constants
from dm_control.manipulation.shared import observations
from dm_control.manipulation.shared import registry
from dm_control.manipulation.shared import robots
from dm_control.manipulation.shared import tags
from dm_control.manipulation.shared import workspaces
from dm_control.utils import rewards
import numpy as np


_ReachWorkspace = collections.namedtuple(
    '_ReachWorkspace', ['target_bbox', 'tcp_bbox', 'arm_offset'])

# Ensures that the props are not touching the table before settling.
_PROP_Z_OFFSET = 0.001

_DUPLO_WORKSPACE = _ReachWorkspace(
    target_bbox=workspaces.BoundingBox(
        lower=(-0.1, -0.1, _PROP_Z_OFFSET),
        upper=(0.1, 0.1, _PROP_Z_OFFSET)),
    tcp_bbox=workspaces.BoundingBox(
        lower=(-0.1, -0.1, 0.2),
        upper=(0.1, 0.1, 0.4)),
    arm_offset=robots.ARM_OFFSET)

_SITE_WORKSPACE = _ReachWorkspace(
    target_bbox=workspaces.BoundingBox(
        lower=(-0.2, -0.2, 0.02),
        upper=(0.2, 0.2, 0.4)),
    tcp_bbox=workspaces.BoundingBox(
        lower=(-0.2, -0.2, 0.02),
        upper=(0.2, 0.2, 0.4)),
    arm_offset=robots.ARM_OFFSET)

_TARGET_RADIUS = 0.05


class Reach(composer.Task):
  """Bring the hand close to a target prop or site."""

  def __init__(
      self, arena, arm, hand, prop, obs_settings, workspace, control_timestep):
    """Initializes a new `Reach` task.

    Args:
      arena: `composer.Entity` instance.
      arm: `robot_base.RobotArm` instance.
      hand: `robot_base.RobotHand` instance.
      prop: `composer.Entity` instance specifying the prop to reach to, or None
        in which case the target is a fixed site whose position is specified by
        the workspace.
      obs_settings: `observations.ObservationSettings` instance.
      workspace: `_ReachWorkspace` specifying the placement of the prop and TCP.
      control_timestep: Float specifying the control timestep in seconds.
    """
    self._arena = arena
    self._arm = arm
    self._hand = hand
    self._arm.attach(self._hand)
    self._arena.attach_offset(self._arm, offset=workspace.arm_offset)
    self.control_timestep = control_timestep
    self._tcp_initializer = initializers.ToolCenterPointInitializer(
        self._hand, self._arm,
        position=distributions.Uniform(*workspace.tcp_bbox),
        quaternion=workspaces.DOWN_QUATERNION)

    # Add custom camera observable.
    self._task_observables = cameras.add_camera_observables(
        arena, obs_settings, cameras.FRONT_CLOSE)

    target_pos_distribution = distributions.Uniform(*workspace.target_bbox)
    self._prop = prop
    if prop:
      # The prop itself is used to visualize the target location.
      self._make_target_site(parent_entity=prop, visible=False)
      self._target = self._arena.add_free_entity(prop)
      self._prop_placer = initializers.PropPlacer(
          props=[prop],
          position=target_pos_distribution,
          quaternion=workspaces.uniform_z_rotation,
          settle_physics=True)
    else:
      self._target = self._make_target_site(parent_entity=arena, visible=True)
      self._target_placer = target_pos_distribution

      obs = observable.MJCFFeature('pos', self._target)
      obs.configure(**obs_settings.prop_pose._asdict())
      self._task_observables['target_position'] = obs

    # Add sites for visualizing the prop and target bounding boxes.
    workspaces.add_bbox_site(
        body=self.root_entity.mjcf_model.worldbody,
        lower=workspace.tcp_bbox.lower, upper=workspace.tcp_bbox.upper,
        rgba=constants.GREEN, name='tcp_spawn_area')
    workspaces.add_bbox_site(
        body=self.root_entity.mjcf_model.worldbody,
        lower=workspace.target_bbox.lower, upper=workspace.target_bbox.upper,
        rgba=constants.BLUE, name='target_spawn_area')

  def _make_target_site(self, parent_entity, visible):
    return workspaces.add_target_site(
        body=parent_entity.mjcf_model.worldbody,
        radius=_TARGET_RADIUS, visible=visible,
        rgba=constants.RED, name='target_site')

  @property
  def root_entity(self):
    return self._arena

  @property
  def arm(self):
    return self._arm

  @property
  def hand(self):
    return self._hand

  @property
  def task_observables(self):
    return self._task_observables

  def get_reward(self, physics):
    hand_pos = physics.bind(self._hand.tool_center_point).xpos
    target_pos = physics.bind(self._target).xpos
    distance = np.linalg.norm(hand_pos - target_pos)
    return rewards.tolerance(
        distance, bounds=(0, _TARGET_RADIUS), margin=_TARGET_RADIUS)

  def initialize_episode(self, physics, random_state):
    self._hand.set_grasp(physics, close_factors=random_state.uniform())
    self._tcp_initializer(physics, random_state)
    if self._prop:
      self._prop_placer(physics, random_state)
    else:
      physics.bind(self._target).pos = (
          self._target_placer(random_state=random_state))


def _reach(obs_settings, use_site):
  """Configure and instantiate a `Reach` task.

  Args:
    obs_settings: An `observations.ObservationSettings` instance.
    use_site: Boolean, if True then the target will be a fixed site, otherwise
      it will be a moveable Duplo brick.

  Returns:
    An instance of `reach.Reach`.
  """
  arena = arenas.Standard()
  arm = robots.make_arm(obs_settings=obs_settings)
  hand = robots.make_hand(obs_settings=obs_settings)
  if use_site:
    workspace = _SITE_WORKSPACE
    prop = None
  else:
    workspace = _DUPLO_WORKSPACE
    prop = props.Duplo(observable_options=observations.make_options(
        obs_settings, observations.FREEPROP_OBSERVABLES))
  task = Reach(arena=arena, arm=arm, hand=hand, prop=prop,
               obs_settings=obs_settings,
               workspace=workspace,
               control_timestep=constants.CONTROL_TIMESTEP)
  return task


@registry.add(tags.FEATURES, tags.EASY)
def reach_duplo_features():
  return _reach(obs_settings=observations.PERFECT_FEATURES, use_site=False)


@registry.add(tags.VISION, tags.EASY)
def reach_duplo_vision():
  return _reach(obs_settings=observations.VISION, use_site=False)


@registry.add(tags.FEATURES, tags.EASY)
def reach_site_features():
  return _reach(obs_settings=observations.PERFECT_FEATURES, use_site=True)


@registry.add(tags.VISION, tags.EASY)
def reach_site_vision():
  return _reach(obs_settings=observations.VISION, use_site=True)



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Fish Domain."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

from dm_control import mujoco
from dm_control.rl import control
from dm_control.suite import base
from dm_control.suite import common
from dm_control.utils import containers
from dm_control.utils import rewards
import numpy as np


_DEFAULT_TIME_LIMIT = 40
_CONTROL_TIMESTEP = .04
_JOINTS = ['tail1',
           'tail_twist',
           'tail2',
           'finright_roll',
           'finright_pitch',
           'finleft_roll',
           'finleft_pitch']
SUITE = containers.TaggedTasks()


def get_model_and_assets():
  """Returns a tuple containing the model XML string and a dict of assets."""
  return common.read_model('fish.xml'), common.ASSETS


@SUITE.add('benchmarking')
def upright(time_limit=_DEFAULT_TIME_LIMIT, random=None,
            environment_kwargs=None):
  """Returns the Fish Upright task."""
  physics = Physics.from_xml_string(*get_model_and_assets())
  task = Upright(random=random)
  environment_kwargs = environment_kwargs or {}
  return control.Environment(
      physics, task, control_timestep=_CONTROL_TIMESTEP, time_limit=time_limit,
      **environment_kwargs)


@SUITE.add('benchmarking')
def swim(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
  """Returns the Fish Swim task."""
  physics = Physics.from_xml_string(*get_model_and_assets())
  task = Swim(random=random)
  environment_kwargs = environment_kwargs or {}
  return control.Environment(
      physics, task, control_timestep=_CONTROL_TIMESTEP, time_limit=time_limit,
      **environment_kwargs)


class Physics(mujoco.Physics):
  """Physics simulation with additional features for the Fish domain."""

  def upright(self):
    """Returns projection from z-axes of torso to the z-axes of worldbody."""
    return self.named.data.xmat['torso', 'zz']

  def torso_velocity(self):
    """Returns velocities and angular velocities of the torso."""
    return self.data.sensordata

  def joint_velocities(self):
    """Returns the joint velocities."""
    return self.named.data.qvel[_JOINTS]

  def joint_angles(self):
    """Returns the joint positions."""
    return self.named.data.qpos[_JOINTS]

  def mouth_to_target(self):
    """Returns a vector, from mouth to target in local coordinate of mouth."""
    data = self.named.data
    mouth_to_target_global = data.geom_xpos['target'] - data.geom_xpos['mouth']
    return mouth_to_target_global.dot(data.geom_xmat['mouth'].reshape(3, 3))


class Upright(base.Task):
  """A Fish `Task` for getting the torso upright with smooth reward."""

  def __init__(self, random=None):
    """Initializes an instance of `Upright`.

    Args:
      random: Either an existing `numpy.random.RandomState` instance, an
        integer seed for creating a new `RandomState`, or None to select a seed
        automatically.
    """
    super(Upright, self).__init__(random=random)

  def initialize_episode(self, physics):
    """Randomizes the tail and fin angles and the orientation of the Fish."""
    quat = self.random.randn(4)
    physics.named.data.qpos['root'][3:7] = quat / np.linalg.norm(quat)
    for joint in _JOINTS:
      physics.named.data.qpos[joint] = self.random.uniform(-.2, .2)
    # Hide the target. It's irrelevant for this task.
    physics.named.model.geom_rgba['target', 3] = 0
    super(Upright, self).initialize_episode(physics)

  def get_observation(self, physics):
    """Returns an observation of joint angles, velocities and uprightness."""
    obs = collections.OrderedDict()
    obs['joint_angles'] = physics.joint_angles()
    obs['upright'] = physics.upright()
    obs['velocity'] = physics.velocity()
    return obs

  def get_reward(self, physics):
    """Returns a smooth reward."""
    return rewards.tolerance(physics.upright(), bounds=(1, 1), margin=1)


class Swim(base.Task):
  """A Fish `Task` for swimming with smooth reward."""

  def __init__(self, random=None):
    """Initializes an instance of `Swim`.

    Args:
      random: Optional, either a `numpy.random.RandomState` instance, an
        integer seed for creating a new `RandomState`, or None to select a seed
        automatically (default).
    """
    super(Swim, self).__init__(random=random)

  def initialize_episode(self, physics):
    """Sets the state of the environment at the start of each episode."""

    quat = self.random.randn(4)
    physics.named.data.qpos['root'][3:7] = quat / np.linalg.norm(quat)
    for joint in _JOINTS:
      physics.named.data.qpos[joint] = self.random.uniform(-.2, .2)
    # Randomize target position.
    physics.named.model.geom_pos['target', 'x'] = self.random.uniform(-.4, .4)
    physics.named.model.geom_pos['target', 'y'] = self.random.uniform(-.4, .4)
    physics.named.model.geom_pos['target', 'z'] = self.random.uniform(.1, .3)
    super(Swim, self).initialize_episode(physics)

  def get_observation(self, physics):
    """Returns an observation of joints, target direction and velocities."""
    obs = collections.OrderedDict()
    obs['joint_angles'] = physics.joint_angles()
    obs['upright'] = physics.upright()
    obs['target'] = physics.mouth_to_target()
    obs['velocity'] = physics.velocity()
    return obs

  def get_reward(self, physics):
    """Returns a smooth reward."""
    radii = physics.named.model.geom_size[['mouth', 'target'], 0].sum()
    in_target = rewards.tolerance(np.linalg.norm(physics.mouth_to_target()),
                                  bounds=(0, radii), margin=2*radii)
    is_upright = 0.5 * (physics.upright() + 1)
    return (7*in_target + is_upright) / 8



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Soft indicator function evaluating whether a number is within bounds."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np

# The value returned by tolerance() at `margin` distance from `bounds` interval.
_DEFAULT_VALUE_AT_MARGIN = 0.1


def _sigmoids(x, value_at_1, sigmoid):
  """Returns 1 when `x` == 0, between 0 and 1 otherwise.

  Args:
    x: A scalar or numpy array.
    value_at_1: A float between 0 and 1 specifying the output when `x` == 1.
    sigmoid: String, choice of sigmoid type.

  Returns:
    A numpy array with values between 0.0 and 1.0.

  Raises:
    ValueError: If not 0 < `value_at_1` < 1, except for `linear`, `cosine` and
      `quadratic` sigmoids which allow `value_at_1` == 0.
    ValueError: If `sigmoid` is of an unknown type.
  """
  if sigmoid in ('cosine', 'linear', 'quadratic'):
    if not 0 <= value_at_1 < 1:
      raise ValueError('`value_at_1` must be nonnegative and smaller than 1, '
                       'got {}.'.format(value_at_1))
  else:
    if not 0 < value_at_1 < 1:
      raise ValueError('`value_at_1` must be strictly between 0 and 1, '
                       'got {}.'.format(value_at_1))

  if sigmoid == 'gaussian':
    scale = np.sqrt(-2 * np.log(value_at_1))
    return np.exp(-0.5 * (x*scale)**2)

  elif sigmoid == 'hyperbolic':
    scale = np.arccosh(1/value_at_1)
    return 1 / np.cosh(x*scale)

  elif sigmoid == 'long_tail':
    scale = np.sqrt(1/value_at_1 - 1)
    return 1 / ((x*scale)**2 + 1)

  elif sigmoid == 'reciprocal':
    scale = 1/value_at_1 - 1
    return 1 / (abs(x)*scale + 1)

  elif sigmoid == 'cosine':
    scale = np.arccos(2*value_at_1 - 1) / np.pi
    scaled_x = x*scale
    return np.where(abs(scaled_x) < 1, (1 + np.cos(np.pi*scaled_x))/2, 0.0)

  elif sigmoid == 'linear':
    scale = 1-value_at_1
    scaled_x = x*scale
    return np.where(abs(scaled_x) < 1, 1 - scaled_x, 0.0)

  elif sigmoid == 'quadratic':
    scale = np.sqrt(1-value_at_1)
    scaled_x = x*scale
    return np.where(abs(scaled_x) < 1, 1 - scaled_x**2, 0.0)

  elif sigmoid == 'tanh_squared':
    scale = np.arctanh(np.sqrt(1-value_at_1))
    return 1 - np.tanh(x*scale)**2

  else:
    raise ValueError('Unknown sigmoid type {!r}.'.format(sigmoid))


def tolerance(x, bounds=(0.0, 0.0), margin=0.0, sigmoid='gaussian',
              value_at_margin=_DEFAULT_VALUE_AT_MARGIN):
  """Returns 1 when `x` falls inside the bounds, between 0 and 1 otherwise.

  Args:
    x: A scalar or numpy array.
    bounds: A tuple of floats specifying inclusive `(lower, upper)` bounds for
      the target interval. These can be infinite if the interval is unbounded
      at one or both ends, or they can be equal to one another if the target
      value is exact.
    margin: Float. Parameter that controls how steeply the output decreases as
      `x` moves out-of-bounds.
      * If `margin == 0` then the output will be 0 for all values of `x`
        outside of `bounds`.
      * If `margin > 0` then the output will decrease sigmoidally with
        increasing distance from the nearest bound.
    sigmoid: String, choice of sigmoid type. Valid values are: 'gaussian',
       'linear', 'hyperbolic', 'long_tail', 'cosine', 'tanh_squared'.
    value_at_margin: A float between 0 and 1 specifying the output value when
      the distance from `x` to the nearest bound is equal to `margin`. Ignored
      if `margin == 0`.

  Returns:
    A float or numpy array with values between 0.0 and 1.0.

  Raises:
    ValueError: If `bounds[0] > bounds[1]`.
    ValueError: If `margin` is negative.
  """
  lower, upper = bounds
  if lower > upper:
    raise ValueError('Lower bound must be <= upper bound.')
  if margin < 0:
    raise ValueError('`margin` must be non-negative.')

  in_bounds = np.logical_and(lower <= x, x <= upper)
  if margin == 0:
    value = np.where(in_bounds, 1.0, 0.0)
  else:
    d = np.where(x < lower, lower - x, x - upper) / margin
    value = np.where(in_bounds, 1.0, _sigmoids(d, value_at_margin, sigmoid))

  return float(value) if np.isscalar(x) else value




# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests for `mjcf.debugging`."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import contextlib
import os
import re
import shutil
import sys

# Internal dependencies.
from absl.testing import absltest
from dm_control import mjcf

from dm_control.mjcf import code_for_debugging_test as test_code
from dm_control.mjcf import debugging
from dm_control.mujoco import wrapper as mujoco_wrapper
import six

ORIGINAL_DEBUG_MODE = debugging.debug_mode()


class DebuggingTest(absltest.TestCase):

  def tearDown(self):
    super(DebuggingTest, self).tearDown()
    if ORIGINAL_DEBUG_MODE:
      debugging.enable_debug_mode()
    else:
      debugging.disable_debug_mode()

  def setup_debug_mode(self, debug_mode_enabled, full_dump_enabled=False):
    if debug_mode_enabled:
      debugging.enable_debug_mode()
    else:
      debugging.disable_debug_mode()
    if full_dump_enabled:
      base_dir = absltest.get_default_test_tmpdir()
      self.dump_dir = os.path.join(base_dir, 'mjcf_debugging_test')
      shutil.rmtree(self.dump_dir, ignore_errors=True)
      os.mkdir(self.dump_dir)
    else:
      self.dump_dir = ''
    debugging.set_full_dump_dir(self.dump_dir)

  def assertStackFromTestCode(self, stack, function_name, line_ref):
    self.assertEqual(stack[-1].function_name, function_name)
    self.assertStartsWith(test_code.__file__, stack[-1].filename)
    line_info = test_code.LINE_REF['.'.join([function_name, line_ref])]
    self.assertEqual(stack[-1].line_number, line_info.line_number)
    self.assertEqual(stack[-1].text, line_info.text)

  @contextlib.contextmanager
  def assertRaisesTestCodeRef(self, line_ref):
    filename, _ = os.path.splitext(test_code.__file__)
    expected_message = (
        filename + '.py:' + str(test_code.LINE_REF[line_ref].line_number))
    print(expected_message)
    with six.assertRaisesRegex(self, mujoco_wrapper.Error, expected_message):
      yield

  def test_get_current_stack_trace(self):
    self.setup_debug_mode(debug_mode_enabled=True)
    stack_trace = debugging.get_current_stack_trace()
    self.assertStartsWith(
        sys.modules[__name__].__file__, stack_trace[-1].filename)
    self.assertEqual(stack_trace[-1].function_name,
                     'test_get_current_stack_trace')
    self.assertEqual(stack_trace[-1].text,
                     'stack_trace = debugging.get_current_stack_trace()')

  def test_disable_debug_mode(self):
    self.setup_debug_mode(debug_mode_enabled=False)
    mjcf_model = test_code.make_valid_model()
    test_code.break_valid_model(mjcf_model)
    self.assertFalse(mjcf_model.get_init_stack())

    my_actuator = mjcf_model.find('actuator', 'my_actuator')
    my_actuator_attrib_stacks = (
        my_actuator.get_last_modified_stacks_for_all_attributes())
    for stack in six.itervalues(my_actuator_attrib_stacks):
      self.assertFalse(stack)

  def test_element_and_attribute_stacks(self):
    self.setup_debug_mode(debug_mode_enabled=True)
    mjcf_model = test_code.make_valid_model()
    test_code.break_valid_model(mjcf_model)

    self.assertStackFromTestCode(mjcf_model.get_init_stack(),
                                 'make_valid_model', 'mjcf_model')

    my_actuator = mjcf_model.find('actuator', 'my_actuator')
    self.assertStackFromTestCode(my_actuator.get_init_stack(),
                                 'make_valid_model', 'my_actuator')

    my_actuator_attrib_stacks = (
        my_actuator.get_last_modified_stacks_for_all_attributes())
    # `name` attribute was assigned at the same time as the element was created.
    self.assertEqual(my_actuator_attrib_stacks['name'],
                     my_actuator.get_init_stack())
    # `joint` attribute was modified later on.
    self.assertStackFromTestCode(my_actuator_attrib_stacks['joint'],
                                 'break_valid_model', 'my_actuator.joint')

  def test_valid_physics(self):
    self.setup_debug_mode(debug_mode_enabled=True)
    mjcf_model = test_code.make_valid_model()
    mjcf.Physics.from_mjcf_model(mjcf_model)  # Should not raise

  def test_physics_error_message_outside_of_debug_mode(self):
    self.setup_debug_mode(debug_mode_enabled=False)
    mjcf_model = test_code.make_broken_model()
    # Make sure that we advertise debug mode if it's currently disabled.
    with six.assertRaisesRegex(self, mujoco_wrapper.Error, '--pymjcf_debug'):
      mjcf.Physics.from_mjcf_model(mjcf_model)

  def test_physics_error_message_in_debug_mode(self):
    self.setup_debug_mode(debug_mode_enabled=True)
    mjcf_model_1 = test_code.make_broken_model()
    with self.assertRaisesTestCodeRef('make_broken_model.my_actuator'):
      mjcf.Physics.from_mjcf_model(mjcf_model_1)
    mjcf_model_2 = test_code.make_valid_model()
    physics = mjcf.Physics.from_mjcf_model(mjcf_model_2)  # Should not raise.
    test_code.break_valid_model(mjcf_model_2)
    with self.assertRaisesTestCodeRef('break_valid_model.my_actuator.joint'):
      physics.reload_from_mjcf_model(mjcf_model_2)

  def test_full_debug_dump(self):
    self.setup_debug_mode(debug_mode_enabled=True, full_dump_enabled=False)
    mjcf_model = test_code.make_valid_model()
    test_code.break_valid_model(mjcf_model)
    # Make sure that we advertise full dump mode if it's currently disabled.
    with six.assertRaisesRegex(self, mujoco_wrapper.Error,
                               '--pymjcf_debug_full_dump_dir'):
      mjcf.Physics.from_mjcf_model(mjcf_model)
    self.setup_debug_mode(debug_mode_enabled=True, full_dump_enabled=True)
    with self.assertRaises(mujoco_wrapper.Error):
      mjcf.Physics.from_mjcf_model(mjcf_model)

    with open(os.path.join(self.dump_dir, 'model.xml')) as f:
      dumped_xml = f.read()
    dumped_xml = [line.strip() for line in dumped_xml.strip().split('\n')]

    xml_line_pattern = re.compile(r'^(.*)<!--pymjcfdebug:(\d+)-->$')
    uninstrumented_pattern = re.compile(r'({})'.format(
        '|'.join([
            r'<mujoco model=".*">',
            r'</mujoco>',
            r'<default class=".*/"/?>',
            r'</default>'
        ])))

    for xml_line in dumped_xml:
      print(xml_line)
      xml_line_match = xml_line_pattern.match(xml_line)
      if not xml_line_match:
        # Only uninstrumented lines are allowed to have no metadata.
        self.assertIsNotNone(uninstrumented_pattern.match(xml_line))
      else:
        xml_element = xml_line_match.group(1)
        debug_id = int(xml_line_match.group(2))
        with open(os.path.join(self.dump_dir, str(debug_id) + '.dump')) as f:
          element_dump = f.read()
        self.assertIn(xml_element, element_dump)

if __name__ == '__main__':
  absltest.main()



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Finger Domain."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

from dm_control import mujoco
from dm_control.rl import control
from dm_control.suite import base
from dm_control.suite import common
from dm_control.suite.utils import randomizers
from dm_control.utils import containers
import numpy as np
from six.moves import range

_DEFAULT_TIME_LIMIT = 20  # (seconds)
_CONTROL_TIMESTEP = .02   # (seconds)
# For TURN tasks, the 'tip' geom needs to enter a spherical target of sizes:
_EASY_TARGET_SIZE = 0.07
_HARD_TARGET_SIZE = 0.03
# Initial spin velocity for the Stop task.
_INITIAL_SPIN_VELOCITY = 100
# Spinning slower than this value (radian/second) is considered stopped.
_STOP_VELOCITY = 1e-6
# Spinning faster than this value (radian/second) is considered spinning.
_SPIN_VELOCITY = 15.0


SUITE = containers.TaggedTasks()


def get_model_and_assets():
  """Returns a tuple containing the model XML string and a dict of assets."""
  return common.read_model('finger.xml'), common.ASSETS


@SUITE.add('benchmarking')
def spin(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
  """Returns the Spin task."""
  physics = Physics.from_xml_string(*get_model_and_assets())
  task = Spin(random=random)
  environment_kwargs = environment_kwargs or {}
  return control.Environment(
      physics, task, time_limit=time_limit, control_timestep=_CONTROL_TIMESTEP,
      **environment_kwargs)


@SUITE.add('benchmarking')
def turn_easy(time_limit=_DEFAULT_TIME_LIMIT, random=None,
              environment_kwargs=None):
  """Returns the easy Turn task."""
  physics = Physics.from_xml_string(*get_model_and_assets())
  task = Turn(target_radius=_EASY_TARGET_SIZE, random=random)
  environment_kwargs = environment_kwargs or {}
  return control.Environment(
      physics, task, time_limit=time_limit, control_timestep=_CONTROL_TIMESTEP,
      **environment_kwargs)


@SUITE.add('benchmarking')
def turn_hard(time_limit=_DEFAULT_TIME_LIMIT, random=None,
              environment_kwargs=None):
  """Returns the hard Turn task."""
  physics = Physics.from_xml_string(*get_model_and_assets())
  task = Turn(target_radius=_HARD_TARGET_SIZE, random=random)
  environment_kwargs = environment_kwargs or {}
  return control.Environment(
      physics, task, time_limit=time_limit, control_timestep=_CONTROL_TIMESTEP,
      **environment_kwargs)


class Physics(mujoco.Physics):
  """Physics simulation with additional features for the Finger domain."""

  def touch(self):
    """Returns logarithmically scaled signals from the two touch sensors."""
    return np.log1p(self.named.data.sensordata[['touchtop', 'touchbottom']])

  def hinge_velocity(self):
    """Returns the velocity of the hinge joint."""
    return self.named.data.sensordata['hinge_velocity']

  def tip_position(self):
    """Returns the (x,z) position of the tip relative to the hinge."""
    return (self.named.data.sensordata['tip'][[0, 2]] -
            self.named.data.sensordata['spinner'][[0, 2]])

  def bounded_position(self):
    """Returns the positions, with the hinge angle replaced by tip position."""
    return np.hstack((self.named.data.sensordata[['proximal', 'distal']],
                      self.tip_position()))

  def velocity(self):
    """Returns the velocities (extracted from sensordata)."""
    return self.named.data.sensordata[['proximal_velocity',
                                       'distal_velocity',
                                       'hinge_velocity']]

  def target_position(self):
    """Returns the (x,z) position of the target relative to the hinge."""
    return (self.named.data.sensordata['target'][[0, 2]] -
            self.named.data.sensordata['spinner'][[0, 2]])

  def to_target(self):
    """Returns the vector from the tip to the target."""
    return self.target_position() - self.tip_position()

  def dist_to_target(self):
    """Returns the signed distance to the target surface, negative is inside."""
    return (np.linalg.norm(self.to_target()) -
            self.named.model.site_size['target', 0])


class Spin(base.Task):
  """A Finger `Task` to spin the stopped body."""

  def __init__(self, random=None):
    """Initializes a new `Spin` instance.

    Args:
      random: Optional, either a `numpy.random.RandomState` instance, an
        integer seed for creating a new `RandomState`, or None to select a seed
        automatically (default).
    """
    super(Spin, self).__init__(random=random)

  def initialize_episode(self, physics):
    physics.named.model.site_rgba['target', 3] = 0
    physics.named.model.site_rgba['tip', 3] = 0
    physics.named.model.dof_damping['hinge'] = .03
    _set_random_joint_angles(physics, self.random)
    super(Spin, self).initialize_episode(physics)

  def get_observation(self, physics):
    """Returns state and touch sensors, and target info."""
    obs = collections.OrderedDict()
    obs['position'] = physics.bounded_position()
    obs['velocity'] = physics.velocity()
    obs['touch'] = physics.touch()
    return obs

  def get_reward(self, physics):
    """Returns a sparse reward."""
    return float(physics.hinge_velocity() <= -_SPIN_VELOCITY)


class Turn(base.Task):
  """A Finger `Task` to turn the body to a target angle."""

  def __init__(self, target_radius, random=None):
    """Initializes a new `Turn` instance.

    Args:
      target_radius: Radius of the target site, which specifies the goal angle.
      random: Optional, either a `numpy.random.RandomState` instance, an
        integer seed for creating a new `RandomState`, or None to select a seed
        automatically (default).
    """
    self._target_radius = target_radius
    super(Turn, self).__init__(random=random)

  def initialize_episode(self, physics):
    target_angle = self.random.uniform(-np.pi, np.pi)
    hinge_x, hinge_z = physics.named.data.xanchor['hinge', ['x', 'z']]
    radius = physics.named.model.geom_size['cap1'].sum()
    target_x = hinge_x + radius * np.sin(target_angle)
    target_z = hinge_z + radius * np.cos(target_angle)
    physics.named.model.site_pos['target', ['x', 'z']] = target_x, target_z
    physics.named.model.site_size['target', 0] = self._target_radius

    _set_random_joint_angles(physics, self.random)

    super(Turn, self).initialize_episode(physics)

  def get_observation(self, physics):
    """Returns state, touch sensors, and target info."""
    obs = collections.OrderedDict()
    obs['position'] = physics.bounded_position()
    obs['velocity'] = physics.velocity()
    obs['touch'] = physics.touch()
    obs['target_position'] = physics.target_position()
    obs['dist_to_target'] = physics.dist_to_target()
    return obs

  def get_reward(self, physics):
    return float(physics.dist_to_target() <= 0)


def _set_random_joint_angles(physics, random, max_attempts=1000):
  """Sets the joints to a random collision-free state."""

  for _ in range(max_attempts):
    randomizers.randomize_limited_and_rotational_joints(physics, random)
    # Check for collisions.
    physics.after_reset()
    if physics.data.ncon == 0:
      break
  else:
    raise RuntimeError('Could not find a collision-free state '
                       'after {} attempts'.format(max_attempts))
